import ast
from dataclasses import dataclass
from typing import ClassVar, cast

from typing_extensions import assert_never

from guppylang.ast_util import get_type, with_loc, with_type
from guppylang.checker.core import Context
from guppylang.checker.errors.generic import ExpectedError, UnsupportedError
from guppylang.checker.errors.type_errors import (
    ArrayComprUnknownSizeError,
    TypeMismatchError,
)
from guppylang.checker.expr_checker import (
    ExprChecker,
    ExprSynthesizer,
    check_num_args,
    check_type_against,
    synthesize_call,
    synthesize_comprehension,
)
from guppylang.definition.custom import (
    CustomCallChecker,
)
from guppylang.definition.struct import CheckedStructDef, RawStructDef
from guppylang.diagnostic import Error, Note
from guppylang.error import GuppyError, GuppyTypeError, InternalGuppyError
from guppylang.nodes import (
    BarrierExpr,
    DesugaredArrayComp,
    DesugaredGeneratorExpr,
    GenericParamValue,
    GlobalCall,
    MakeIter,
    PanicExpr,
    ResultExpr,
)
from guppylang.tys.arg import ConstArg, TypeArg
from guppylang.tys.builtin import (
    array_type,
    array_type_def,
    bool_type,
    get_element_type,
    get_iter_size,
    int_type,
    is_array_type,
    is_bool_type,
    is_sized_iter_type,
    nat_type,
    sized_iter_type,
    string_type,
)
from guppylang.tys.const import Const, ConstValue
from guppylang.tys.subst import Subst
from guppylang.tys.ty import (
    FuncInput,
    FunctionType,
    InputFlags,
    NoneType,
    NumericType,
    StructType,
    Type,
    unify,
)


class ReversingChecker(CustomCallChecker):
    """Call checker for reverse arithmetic methods.

    For examples, turns a call to `__radd__` into a call to `__add__` with reversed
    arguments.
    """

    def parse_name(self) -> str:
        # Must be a dunder method
        assert self.func.name.startswith("__")
        assert self.func.name.endswith("__")
        name = self.func.name[2:-2]
        # Remove the `r`
        assert name.startswith("r")
        return f"__{name[1:]}__"

    def synthesize(self, args: list[ast.expr]) -> tuple[ast.expr, Type]:
        [self_arg, other_arg] = args
        self_arg, self_ty = ExprSynthesizer(self.ctx).synthesize(self_arg)
        f = self.ctx.globals.get_instance_func(self_ty, self.parse_name())
        assert f is not None
        return f.synthesize_call([other_arg, self_arg], self.node, self.ctx)


class UnsupportedChecker(CustomCallChecker):
    """Call checker for Python builtin functions that are not available in Guppy.

    Gives the uses a nicer error message when they try to use an unsupported feature.
    """

    def synthesize(self, args: list[ast.expr]) -> tuple[ast.expr, Type]:
        err = UnsupportedError(
            self.node, f"Builtin method `{self.func.name}`", singular=True
        )
        raise GuppyError(err)

    def check(self, args: list[ast.expr], ty: Type) -> tuple[ast.expr, Subst]:
        err = UnsupportedError(
            self.node, f"Builtin method `{self.func.name}`", singular=True
        )
        raise GuppyError(err)


class DunderChecker(CustomCallChecker):
    """Call checker for builtin functions that call out to dunder instance methods"""

    dunder_name: str
    num_args: int

    def __init__(self, dunder_name: str, num_args: int = 1):
        assert num_args > 0
        self.dunder_name = dunder_name
        self.num_args = num_args

    def synthesize(self, args: list[ast.expr]) -> tuple[ast.expr, Type]:
        check_num_args(self.num_args, len(args), self.node)
        fst, *rest = args
        return ExprSynthesizer(self.ctx).synthesize_instance_func(
            fst,
            rest,
            self.dunder_name,
            f"a valid argument to `{self.func.name}`",
            give_reason=True,
        )


class CallableChecker(CustomCallChecker):
    """Call checker for the builtin `callable` function"""

    def synthesize(self, args: list[ast.expr]) -> tuple[ast.expr, Type]:
        check_num_args(1, len(args), self.node)
        [arg] = args
        arg, ty = ExprSynthesizer(self.ctx).synthesize(arg)
        is_callable = (
            isinstance(ty, FunctionType)
            or self.ctx.globals.get_instance_func(ty, "__call__") is not None
        )
        const = with_loc(self.node, ast.Constant(value=is_callable))
        return const, bool_type()


class ArrayCopyChecker(CustomCallChecker):
    """Function call checker for the `array.copy` function."""

    @dataclass(frozen=True)
    class NonCopyableElementsError(Error):
        title: ClassVar[str] = "Non-copyable elements"
        span_label: ClassVar[str] = "Elements of type `{ty}` cannot be copied."
        ty: Type

        @dataclass(frozen=True)
        class Explanation(Note):
            message: ClassVar[str] = "Only arrays with copyable elements can be copied"

    def synthesize(self, args: list[ast.expr]) -> tuple[ast.expr, Type]:
        # First, check if we're trying to copy a non-copyable element type to give a
        # nicer error message. Then, do the full `synthesize_call` type check
        if len(args) == 1:
            args[0], array_ty = ExprSynthesizer(self.ctx).synthesize(args[0])
            if is_array_type(array_ty):
                elem_ty = get_element_type(array_ty)
                if not elem_ty.copyable:
                    err = ArrayCopyChecker.NonCopyableElementsError(self.node, elem_ty)
                    err.add_sub_diagnostic(
                        ArrayCopyChecker.NonCopyableElementsError.Explanation(None)
                    )
                    raise GuppyTypeError(err)
        [array_arg], _, inst = synthesize_call(self.func.ty, args, self.node, self.ctx)
        node = GlobalCall(def_id=self.func.id, args=[array_arg], type_args=inst)
        return with_loc(self.node, node), get_type(array_arg)


class NewArrayChecker(CustomCallChecker):
    """Function call checker for the `array.__new__` function."""

    @dataclass(frozen=True)
    class InferenceError(Error):
        title: ClassVar[str] = "Cannot infer type"
        span_label: ClassVar[str] = "Cannot infer the type of this array"

        @dataclass(frozen=True)
        class Suggestion(Note):
            message: ClassVar[str] = (
                "Consider adding a type annotation: `x: array[???] = ...`"
            )

    def synthesize(self, args: list[ast.expr]) -> tuple[ast.expr, Type]:
        match args:
            case []:
                err = NewArrayChecker.InferenceError(self.node)
                err.add_sub_diagnostic(NewArrayChecker.InferenceError.Suggestion(None))
                raise GuppyTypeError(err)
            # Either an array comprehension
            case [DesugaredGeneratorExpr() as compr]:
                return self.synthesize_array_comprehension(compr)
            # Or a list of array elements
            case [fst, *rest]:
                fst, ty = ExprSynthesizer(self.ctx).synthesize(fst)
                checker = ExprChecker(self.ctx)
                for i in range(len(rest)):
                    rest[i], subst = checker.check(rest[i], ty)
                    assert len(subst) == 0, "Array element type is closed"
                result_ty = array_type(ty, len(args))
                call = GlobalCall(
                    def_id=self.func.id, args=[fst, *rest], type_args=result_ty.args
                )
                return with_loc(self.node, call), result_ty
            case args:
                return assert_never(args)  # type: ignore[arg-type]

    def check(self, args: list[ast.expr], ty: Type) -> tuple[ast.expr, Subst]:
        if not is_array_type(ty):
            dummy_array_ty = array_type_def.check_instantiate(
                [p.to_existential()[0] for p in array_type_def.params],
                self.ctx.globals,
                self.node,
            )
            raise GuppyTypeError(TypeMismatchError(self.node, ty, dummy_array_ty))
        subst: Subst = {}
        match ty.args:
            case [TypeArg(ty=elem_ty), ConstArg(length)]:
                match args:
                    # Either an array comprehension
                    case [DesugaredGeneratorExpr() as compr]:
                        # TODO: We could use the type information to infer some stuff
                        #  in the comprehension
                        arr_compr, res_ty = self.synthesize_array_comprehension(compr)
                        arr_compr = with_loc(self.node, arr_compr)
                        arr_compr, subst, _ = check_type_against(
                            res_ty, ty, arr_compr, self.ctx
                        )
                        return arr_compr, subst
                    # Or a list of array elements
                    case args:
                        checker = ExprChecker(self.ctx)
                        for i in range(len(args)):
                            args[i], s = checker.check(
                                args[i], elem_ty.substitute(subst)
                            )
                            subst |= s
                        ls = unify(length, ConstValue(nat_type(), len(args)), {})
                        if ls is None:
                            raise GuppyTypeError(
                                TypeMismatchError(
                                    self.node, ty, array_type(elem_ty, len(args))
                                )
                            )
                        subst |= ls
                        call = GlobalCall(
                            def_id=self.func.id, args=args, type_args=ty.args
                        )
                        return with_loc(self.node, call), subst
            case type_args:
                raise InternalGuppyError(f"Invalid array type args: {type_args}")

    def synthesize_array_comprehension(
        self, compr: DesugaredGeneratorExpr
    ) -> tuple[DesugaredArrayComp, Type]:
        # Array comprehensions require a static size. To keep things simple, we'll only
        # allow a single generator for now, so we don't have to reason about products
        # of iterator sizes.
        if len(compr.generators) > 1:
            # Individual generator objects unfortunately don't have a span in Python's
            # AST, so we have to use the whole expression span
            raise GuppyError(UnsupportedError(compr, "Nested array comprehensions"))
        [gen] = compr.generators
        # Similarly, dynamic if guards are not allowed
        if gen.ifs:
            err = ArrayComprUnknownSizeError(compr)
            err.add_sub_diagnostic(ArrayComprUnknownSizeError.IfGuard(gen.ifs[0]))
            raise GuppyError(err)
        # Extract the iterator size
        match gen.iter_assign:
            case ast.Assign(value=MakeIter() as make_iter):
                sized_make_iter = MakeIter(
                    make_iter.value, make_iter.origin_node, unwrap_size_hint=False
                )
                _, iter_ty = ExprSynthesizer(self.ctx).synthesize(sized_make_iter)
                # The iterator must have a static size hint
                if not is_sized_iter_type(iter_ty):
                    err = ArrayComprUnknownSizeError(compr)
                    err.add_sub_diagnostic(
                        ArrayComprUnknownSizeError.DynamicIterator(make_iter)
                    )
                    raise GuppyError(err)
                size = get_iter_size(iter_ty)
            case _:
                raise InternalGuppyError("Invalid iterator assign statement")
        # Finally, type check the comprehension
        [gen], elt, elt_ty = synthesize_comprehension(compr, [gen], compr.elt, self.ctx)
        array_compr = DesugaredArrayComp(
            elt=elt, generator=gen, length=size, elt_ty=elt_ty
        )
        return with_loc(compr, array_compr), array_type(elt_ty, size)


#: Maximum length of a tag in the `result` function.
TAG_MAX_LEN = 200


class ResultChecker(CustomCallChecker):
    """Call checker for the `result` function."""

    @dataclass(frozen=True)
    class InvalidError(Error):
        title: ClassVar[str] = "Invalid Result"
        span_label: ClassVar[str] = "Expression of type `{ty}` is not a valid result."
        ty: Type

        @dataclass(frozen=True)
        class Explanation(Note):
            message: ClassVar[str] = (
                "Only numeric values or arrays thereof are allowed as results"
            )

    @dataclass(frozen=True)
    class TooLongError(Error):
        title: ClassVar[str] = "Tag too long"
        span_label: ClassVar[str] = "Result tag is too long"

        @dataclass(frozen=True)
        class Hint(Note):
            message: ClassVar[str] = f"Result tags are limited to {TAG_MAX_LEN} bytes"

    def synthesize(self, args: list[ast.expr]) -> tuple[ast.expr, Type]:
        check_num_args(2, len(args), self.node)
        [tag, value] = args
        tag, _ = ExprChecker(self.ctx).check(tag, string_type())
        if not isinstance(tag, ast.Constant) or not isinstance(tag.value, str):
            raise GuppyTypeError(ExpectedError(tag, "a string literal"))
        if len(tag.value.encode("utf-8")) > TAG_MAX_LEN:
            err: Error = ResultChecker.TooLongError(tag)
            err.add_sub_diagnostic(ResultChecker.TooLongError.Hint(None))
            raise GuppyTypeError(err)
        value, ty = ExprSynthesizer(self.ctx).synthesize(value)
        # We only allow numeric values or vectors of numeric values
        err = ResultChecker.InvalidError(value, ty)
        err.add_sub_diagnostic(ResultChecker.InvalidError.Explanation(None))
        if self._is_numeric_or_bool_type(ty):
            base_ty = ty
            array_len: Const | None = None
        elif is_array_type(ty):
            [ty_arg, len_arg] = ty.args
            assert isinstance(ty_arg, TypeArg)
            assert isinstance(len_arg, ConstArg)
            if not self._is_numeric_or_bool_type(ty_arg.ty):
                raise GuppyError(err)
            _base_ty = ty_arg.ty
            _array_len = len_arg.const
            # See https://github.com/CQCL/guppylang/issues/631
            raise GuppyError(UnsupportedError(value, "Array results"))
        else:
            raise GuppyError(err)
        node = ResultExpr(value, base_ty, array_len, tag.value)
        return with_loc(self.node, node), NoneType()

    def check(self, args: list[ast.expr], ty: Type) -> tuple[ast.expr, Subst]:
        expr, res_ty = self.synthesize(args)
        expr, subst, _ = check_type_against(res_ty, ty, expr, self.ctx)
        return expr, subst

    @staticmethod
    def _is_numeric_or_bool_type(ty: Type) -> bool:
        return isinstance(ty, NumericType) or is_bool_type(ty)


class PanicChecker(CustomCallChecker):
    """Call checker for the `panic` function."""

    @dataclass(frozen=True)
    class NoMessageError(Error):
        title: ClassVar[str] = "No panic message"
        span_label: ClassVar[str] = "Missing message argument to panic call"

        @dataclass(frozen=True)
        class Suggestion(Note):
            message: ClassVar[str] = 'Add a message: `panic("message")`'

    def synthesize(self, args: list[ast.expr]) -> tuple[ast.expr, Type]:
        match args:
            case []:
                err = PanicChecker.NoMessageError(self.node)
                err.add_sub_diagnostic(PanicChecker.NoMessageError.Suggestion(None))
                raise GuppyTypeError(err)
            case [msg, *rest]:
                msg, _ = ExprChecker(self.ctx).check(msg, string_type())
                if not isinstance(msg, ast.Constant) or not isinstance(msg.value, str):
                    raise GuppyTypeError(ExpectedError(msg, "a string literal"))

                vals = [ExprSynthesizer(self.ctx).synthesize(val)[0] for val in rest]
                node = PanicExpr(msg.value, vals)
                return with_loc(self.node, node), NoneType()
            case args:
                return assert_never(args)  # type: ignore[arg-type]

    def check(self, args: list[ast.expr], ty: Type) -> tuple[ast.expr, Subst]:
        # Panic may return any type, so we don't have to check anything. Consequently
        # we also can't infer anything in the expected type, so we always return an
        # empty substitution
        expr, _ = self.synthesize(args)
        return expr, {}


class RangeChecker(CustomCallChecker):
    """Call checker for the `range` function."""

    def synthesize(self, args: list[ast.expr]) -> tuple[ast.expr, Type]:
        check_num_args(1, len(args), self.node)
        [stop] = args
        stop_checked, _ = ExprChecker(self.ctx).check(stop, int_type(), "argument")
        range_iter, range_ty = self.make_range(stop_checked)
        # Check if `stop` is a statically known value. Note that we need to do this on
        # the original `stop` instead of `stop_checked` to avoid any previously inserted
        # `int` coercions.
        if (static_stop := self.check_static(stop)) is not None:
            return to_sized_iter(range_iter, range_ty, static_stop, self.ctx)
        return range_iter, range_ty

    def check_static(self, stop: ast.expr) -> "int | Const | None":
        stop, _ = ExprSynthesizer(self.ctx).synthesize(stop, allow_free_vars=True)
        if isinstance(stop, ast.Constant) and isinstance(stop.value, int):
            return stop.value
        if isinstance(stop, GenericParamValue) and stop.param.ty == nat_type():
            return stop.param.to_bound().const
        return None

    def range_ty(self) -> StructType:
        from guppylang.std.builtins import Range

        def_id = cast(RawStructDef, Range).id
        range_type_def = self.ctx.globals.defs[def_id]
        assert isinstance(range_type_def, CheckedStructDef)
        return StructType([], range_type_def)

    def make_range(self, stop: ast.expr) -> tuple[ast.expr, Type]:
        make_range = self.ctx.globals.get_instance_func(self.range_ty(), "__new__")
        assert make_range is not None
        start = with_type(int_type(), with_loc(self.node, ast.Constant(value=0)))
        return make_range.synthesize_call([start, stop], self.node, self.ctx)


def to_sized_iter(
    iterator: ast.expr, range_ty: Type, size: "int | Const", ctx: Context
) -> tuple[ast.expr, Type]:
    """Adds a static size annotation to an iterator."""
    sized_iter_ty = sized_iter_type(range_ty, size)
    make_sized_iter = ctx.globals.get_instance_func(sized_iter_ty, "__new__")
    assert make_sized_iter is not None
    sized_iter, _ = make_sized_iter.check_call([iterator], sized_iter_ty, iterator, ctx)
    return sized_iter, sized_iter_ty


class BarrierChecker(CustomCallChecker):
    """Call checker for the `barrier` function."""

    def synthesize(self, args: list[ast.expr]) -> tuple[ast.expr, Type]:
        tys = [ExprSynthesizer(self.ctx).synthesize(val)[1] for val in args]
        func_ty = FunctionType(
            [FuncInput(t, InputFlags.Inout) for t in tys],
            NoneType(),
        )
        args, ret_ty, inst = synthesize_call(func_ty, args, self.node, self.ctx)
        assert len(inst) == 0, "func_ty is not generic"
        node = BarrierExpr(args=args, func_ty=func_ty)
        return with_loc(self.node, node), ret_ty



"""Snapshot tests for diagnostics rendering"""

from dataclasses import dataclass
from pathlib import Path
from typing import ClassVar

from guppylang.diagnostic import (
    Diagnostic,
    DiagnosticsRenderer,
    Error,
    Help,
    Note,
)
from guppylang.span import Loc, SourceMap, Span

file = "<unknown>"


def run_test(source: str, diagnostic: Diagnostic, snapshot, request):
    sources = SourceMap()
    sources.add_file(file, source)

    renderer = DiagnosticsRenderer(sources)
    renderer.render_diagnostic(diagnostic)
    out = "\n".join(renderer.buffer)

    snapshot.snapshot_dir = str(Path(request.fspath).parent / "snapshots")
    snapshot.assert_match(out, f"{request.node.name}.txt")


@dataclass(frozen=True)
class MyError(Error):
    title: ClassVar[str] = "Can't compare apples with oranges"
    span_label: ClassVar[str] = "Comparison attempted here"


def test_only_title(snapshot, request):
    @dataclass(frozen=True)
    class MyDiagnostic(Error):
        title: ClassVar[str] = "Can't compare apples with oranges"

    source = ""
    diagnostic = MyDiagnostic(None)
    run_test(source, diagnostic, snapshot, request)


def test_only_message(snapshot, request):
    @dataclass(frozen=True)
    class MyDiagnostic(Error):
        title: ClassVar[str] = "Can't compare apples with oranges"
        message: ClassVar[str] = (
            "Can't compare apples with oranges. Please refer to Barone (BMJ, 2000), "
            "https://doi.org/10.1136%2Fbmj.321.7276.1569 for further details."
        )

    source = ""
    diagnostic = MyDiagnostic(None)
    run_test(source, diagnostic, snapshot, request)


def test_only_label(snapshot, request):
    source = "apple == orange"
    span = Span(Loc(file, 1, 6), Loc(file, 1, 8))
    diagnostic = MyError(span)
    run_test(source, diagnostic, snapshot, request)


def test_message_with_span(snapshot, request):
    @dataclass(frozen=True)
    class MyDiagnostic(Error):
        title: ClassVar[str] = "Can't compare apples with oranges"
        message: ClassVar[str] = (
            "Please refer to Barone (BMJ, 2000), "
            "https://doi.org/10.1136%2Fbmj.321.7276.1569 for further details."
        )

    source = "apple == orange"
    span = Span(Loc(file, 1, 6), Loc(file, 1, 8))
    diagnostic = MyDiagnostic(span)
    run_test(source, diagnostic, snapshot, request)


def test_three_labels_formatted(snapshot, request):
    @dataclass(frozen=True)
    class MySubDiagnostic(Note):
        thing: str
        span_label: ClassVar[str] = "This is an {thing}"

    source = "apple == orange"
    span = Span(Loc(file, 1, 6), Loc(file, 1, 8))
    span_apple = Span(Loc(file, 1, 0), Loc(file, 1, 5))
    span_orange = Span(Loc(file, 1, 9), Loc(file, 1, 15))
    diagnostic = MyError(span)
    diagnostic.add_sub_diagnostic(MySubDiagnostic(span_apple, "apple"))
    diagnostic.add_sub_diagnostic(MySubDiagnostic(span_orange, "orange"))
    run_test(source, diagnostic, snapshot, request)


def test_advanced_formatting(snapshot, request):
    @dataclass(frozen=True)
    class MyDiagnostic(Error):
        title: ClassVar[str] = "Can't compare apples with oranges"
        span_label: ClassVar[str] = "This is an {a}{pp}{le}"
        a: str

        @property
        def pp(self) -> str:
            return "pp"

        @property
        def le(self) -> str:
            return "le"

    @dataclass(frozen=True)
    class MySubDiagnostic(Note):
        span_label: ClassVar[str] = "This is not an {a}{pp}{p}{le}"
        p: str

        @property
        def pp(self) -> str:
            return "p"

    source = "apple == orange"
    span_apple = Span(Loc(file, 1, 0), Loc(file, 1, 5))
    span_orange = Span(Loc(file, 1, 9), Loc(file, 1, 15))
    diagnostic = MyDiagnostic(span_apple, "a")
    diagnostic.add_sub_diagnostic(MySubDiagnostic(span_orange, "p"))
    run_test(source, diagnostic, snapshot, request)


def test_long_label(snapshot, request):
    @dataclass(frozen=True)
    class MyDiagnostic(Error):
        title: ClassVar[str] = "Can't compare apples with oranges"
        span_label: ClassVar[str] = "Comparison attempted here. " * 20

    source = "apple == orange"
    span = Span(Loc(file, 1, 6), Loc(file, 1, 8))
    diagnostic = MyDiagnostic(span)
    run_test(source, diagnostic, snapshot, request)


def test_help(snapshot, request):
    @dataclass(frozen=True)
    class MySubDiagnostic(Help):
        message: ClassVar[str] = "Have you tried peeling the orange?"

    source = "apple == orange"
    span = Span(Loc(file, 1, 6), Loc(file, 1, 8))
    diagnostic = MyError(span)
    diagnostic.add_sub_diagnostic(MySubDiagnostic(None))
    run_test(source, diagnostic, snapshot, request)


def test_note(snapshot, request):
    @dataclass(frozen=True)
    class MySubDiagnostic(Note):
        message: ClassVar[str] = "Stop trying, this is a fruitless endeavor"

    source = "apple == orange"
    span = Span(Loc(file, 1, 6), Loc(file, 1, 8))
    diagnostic = MyError(span)
    diagnostic.add_sub_diagnostic(MySubDiagnostic(None))
    run_test(source, diagnostic, snapshot, request)


def test_context(snapshot, request):
    source = "super_apple := apple ** 2\nlemon := orange - apple\napple == orange"
    span = Span(Loc(file, 3, 6), Loc(file, 3, 8))
    diagnostic = MyError(span)
    run_test(source, diagnostic, snapshot, request)


def test_justify_line_number(snapshot, request):
    source = "foo\n" * 99 + "apple == orange"
    span = Span(Loc(file, 100, 6), Loc(file, 100, 8))
    diagnostic = MyError(span)
    run_test(source, diagnostic, snapshot, request)


def test_two_spans_different_lineno_lens(snapshot, request):
    @dataclass(frozen=True)
    class MySubDiagnostic(Note):
        span_label: ClassVar[str] = "This is an apple"

    source = 99 * "apple == orange\n" + "This apple is on another line"
    span1 = Span(Loc(file, 99, 6), Loc(file, 99, 8))
    span2 = Span(Loc(file, 100, 5), Loc(file, 100, 10))
    diagnostic = MyError(span1)
    diagnostic.add_sub_diagnostic(MySubDiagnostic(span2))
    run_test(source, diagnostic, snapshot, request)


def test_indented(snapshot, request):
    source = " " * 50 + "super_apple := apple ** 2\n"
    source += " " * 50 + "    lemon := orange - apple\n"
    source += " " * 50 + "        apple == orange"
    span = Span(Loc(file, 3, 50 + 8 + 6), Loc(file, 3, 50 + 8 + 8))
    diagnostic = MyError(span)
    run_test(source, diagnostic, snapshot, request)


def test_two_line_span(snapshot, request):
    source = "apple.compare(\n     orange) == EQUAL"
    span = Span(Loc(file, 1, 5), Loc(file, 2, 12))
    diagnostic = MyError(span)
    run_test(source, diagnostic, snapshot, request)


def test_three_line_span(snapshot, request):
    source = "apple.compare(\n     orange\n) == EQUAL"
    span = Span(Loc(file, 1, 5), Loc(file, 3, 1))
    diagnostic = MyError(span)
    run_test(source, diagnostic, snapshot, request)


def test_message_new_line(snapshot, request):
    @dataclass(frozen=True)
    class MyError(Error):
        message: ClassVar[str] = "Apple " * 20 + "\n\n" + "Orange " * 30

    diagnostic = MyError(None)
    run_test("", diagnostic, snapshot, request)



from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.builtins import owned
from guppylang.std.quantum import qubit

module = GuppyModule("test")
module.load(qubit)


@guppy.declare(module)
def foo(q: qubit, x: int) -> int: ...


@guppy(module)
def test(q: qubit @owned) -> tuple[int, qubit]:
    # This doesn't work since arguments are evaluated from left to right
    return foo(q, foo(q, foo(q, 0))), q


module.compile()



import ast
from collections.abc import Sequence

from guppylang.ast_util import (
    AstNode,
    set_location_from,
    shift_loc,
)
from guppylang.cfg.builder import is_py_expression
from guppylang.checker.core import Context, Globals, Locals
from guppylang.checker.errors.generic import ExpectedError
from guppylang.definition.common import Definition
from guppylang.definition.module import ModuleDef
from guppylang.definition.parameter import ParamDef
from guppylang.definition.ty import TypeDef
from guppylang.error import GuppyError
from guppylang.tys.arg import Argument, ConstArg, TypeArg
from guppylang.tys.builtin import CallableTypeDef
from guppylang.tys.const import ConstValue
from guppylang.tys.errors import (
    FlagNotAllowedError,
    FreeTypeVarError,
    HigherKindedTypeVarError,
    IllegalPyTypeArgError,
    InvalidCallableTypeError,
    InvalidFlagError,
    InvalidTypeArgError,
    InvalidTypeError,
    ModuleMemberNotFoundError,
    NonLinearOwnedError,
)
from guppylang.tys.param import Parameter, TypeParam
from guppylang.tys.ty import (
    FuncInput,
    FunctionType,
    InputFlags,
    NoneType,
    NumericType,
    TupleType,
    Type,
)


def arg_from_ast(
    node: AstNode,
    globals: Globals,
    param_var_mapping: dict[str, Parameter],
    allow_free_vars: bool = False,
) -> Argument:
    """Turns an AST expression into an argument."""
    # A single (possibly qualified) identifier
    if defn := _try_parse_defn(node, globals):
        return _arg_from_instantiated_defn(
            defn, [], globals, node, param_var_mapping, allow_free_vars
        )

    # A parametrised type, e.g. `list[??]`
    if isinstance(node, ast.Subscript) and (
        defn := _try_parse_defn(node.value, globals)
    ):
        arg_nodes = (
            node.slice.elts if isinstance(node.slice, ast.Tuple) else [node.slice]
        )
        return _arg_from_instantiated_defn(
            defn, arg_nodes, globals, node, param_var_mapping, allow_free_vars
        )

    # We allow tuple types to be written as `(int, bool)`
    if isinstance(node, ast.Tuple):
        ty = TupleType(
            [
                type_from_ast(el, globals, param_var_mapping, allow_free_vars)
                for el in node.elts
            ]
        )
        return TypeArg(ty)

    # `None` is represented as a `ast.Constant` node with value `None`
    if isinstance(node, ast.Constant) and node.value is None:
        return TypeArg(NoneType())

    # Integer literals are turned into nat args since these are the only ones we support
    # right now.
    # TODO: Once we also have int args etc, we need proper inference logic here
    if isinstance(node, ast.Constant) and isinstance(node.value, int):
        # Fun fact: int ast.Constant values are never negative since e.g. `-5` is a
        # `ast.UnaryOp` negation of a `ast.Constant(5)`
        assert node.value >= 0
        nat_ty = NumericType(NumericType.Kind.Nat)
        return ConstArg(ConstValue(nat_ty, node.value))

    # Py-expressions can also be used to specify static numbers
    if py_expr := is_py_expression(node):
        from guppylang.checker.expr_checker import eval_py_expr

        v = eval_py_expr(py_expr, Context(globals, Locals({}), {}))
        if isinstance(v, int):
            nat_ty = NumericType(NumericType.Kind.Nat)
            return ConstArg(ConstValue(nat_ty, v))
        else:
            raise GuppyError(IllegalPyTypeArgError(node, v))

    # Finally, we also support delayed annotations in strings
    if isinstance(node, ast.Constant) and isinstance(node.value, str):
        node = _parse_delayed_annotation(node.value, node)
        return arg_from_ast(node, globals, param_var_mapping, allow_free_vars)

    raise GuppyError(InvalidTypeArgError(node))


def _try_parse_defn(node: AstNode, globals: Globals) -> Definition | None:
    """Tries to parse a (possibly qualified) name into a global definition."""
    from guppylang.checker.cfg_checker import VarNotDefinedError

    match node:
        case ast.Name(id=x):
            if x not in globals:
                raise GuppyError(VarNotDefinedError(node, x))
            return globals[x]
        case ast.Attribute(value=ast.Name(id=module_name) as value, attr=x):
            if module_name not in globals:
                raise GuppyError(VarNotDefinedError(value, module_name))
            module_def = globals[module_name]
            if not isinstance(module_def, ModuleDef):
                err = ExpectedError(
                    value,
                    "a module",
                    got=f"{module_def.description} `{module_def.name}`",
                )
                raise GuppyError(err)
            if x not in module_def.globals:
                raise GuppyError(ModuleMemberNotFoundError(node, module_def.name, x))
            return module_def.globals[x]
        case _:
            return None


def _arg_from_instantiated_defn(
    defn: Definition,
    arg_nodes: list[ast.expr],
    globals: Globals,
    node: AstNode,
    param_var_mapping: dict[str, Parameter],
    allow_free_vars: bool = False,
) -> Argument:
    """Parses a globals definition with type args into an argument."""
    match defn:
        # Special case for the `Callable` type
        case CallableTypeDef():
            return TypeArg(
                _parse_callable_type(
                    arg_nodes, node, globals, param_var_mapping, allow_free_vars
                )
            )
        # Either a defined type (e.g. `int`, `bool`, ...)
        case TypeDef() as defn:
            args = [
                arg_from_ast(arg_node, globals, param_var_mapping, allow_free_vars)
                for arg_node in arg_nodes
            ]
            ty = defn.check_instantiate(args, globals, node)
            return TypeArg(ty)
        # Or a parameter (e.g. `T`, `n`, ...)
        case ParamDef() as defn:
            # We don't allow parametrised variables like `T[int]`
            if arg_nodes:
                raise GuppyError(HigherKindedTypeVarError(node, defn))
            if defn.name not in param_var_mapping:
                if allow_free_vars:
                    param_var_mapping[defn.name] = defn.to_param(len(param_var_mapping))
                else:
                    raise GuppyError(FreeTypeVarError(node, defn))
            return param_var_mapping[defn.name].to_bound()
        case defn:
            err = ExpectedError(node, "a type", got=f"{defn.description} `{defn.name}`")
            raise GuppyError(err)


def _parse_delayed_annotation(ast_str: str, node: ast.Constant) -> ast.expr:
    """Parses a delayed type annotation in a string."""
    try:
        [stmt] = ast.parse(ast_str).body
        if not isinstance(stmt, ast.Expr):
            raise GuppyError(InvalidTypeError(node))
        set_location_from(stmt, loc=node)
        shift_loc(
            stmt,
            delta_lineno=node.lineno - 1,  # -1 since lines start at 1
            delta_col_offset=node.col_offset + 1,  # +1 to remove the `"`
        )
    except (SyntaxError, ValueError):
        raise GuppyError(InvalidTypeError(node)) from None
    else:
        return stmt.value


def _parse_callable_type(
    args: list[ast.expr],
    loc: AstNode,
    globals: Globals,
    param_var_mapping: dict[str, Parameter],
    allow_free_vars: bool = False,
) -> FunctionType:
    """Helper function to parse a `Callable[[<arguments>], <return type>]` type."""
    err = InvalidCallableTypeError(loc)
    if len(args) != 2:
        raise GuppyError(err)
    [inputs, output] = args
    if not isinstance(inputs, ast.List):
        raise GuppyError(err)
    inouts, output = parse_function_io_types(
        inputs.elts, output, loc, globals, param_var_mapping, allow_free_vars
    )
    return FunctionType(inouts, output)


def parse_function_io_types(
    input_nodes: list[ast.expr],
    output_node: ast.expr,
    loc: AstNode,
    globals: Globals,
    param_var_mapping: dict[str, Parameter],
    allow_free_vars: bool = False,
) -> tuple[list[FuncInput], Type]:
    """Parses the inputs and output types of a function type.

    This function takes care of parsing annotations and any related checks.

    Returns the parsed input and output types.
    """
    inputs = []
    for inp in input_nodes:
        ty, flags = type_with_flags_from_ast(
            inp, globals, param_var_mapping, allow_free_vars
        )
        if InputFlags.Owned in flags and ty.copyable:
            raise GuppyError(NonLinearOwnedError(loc, ty))
        if not ty.copyable and InputFlags.Owned not in flags:
            flags |= InputFlags.Inout

        inputs.append(FuncInput(ty, flags))
    output = type_from_ast(output_node, globals, param_var_mapping, allow_free_vars)
    return inputs, output


_type_param = TypeParam(0, "T", False, False)


def type_with_flags_from_ast(
    node: AstNode,
    globals: Globals,
    param_var_mapping: dict[str, Parameter],
    allow_free_vars: bool = False,
) -> tuple[Type, InputFlags]:
    if isinstance(node, ast.BinOp) and isinstance(node.op, ast.MatMult):
        ty, flags = type_with_flags_from_ast(
            node.left, globals, param_var_mapping, allow_free_vars
        )
        match node.right:
            case ast.Name(id="owned"):
                if ty.copyable:
                    raise GuppyError(NonLinearOwnedError(node.right, ty))
                flags |= InputFlags.Owned
            case _:
                raise GuppyError(InvalidFlagError(node.right))
        return ty, flags
    # We also need to handle the case that this could be a delayed string annotation
    elif isinstance(node, ast.Constant) and isinstance(node.value, str):
        node = _parse_delayed_annotation(node.value, node)
        return type_with_flags_from_ast(
            node, globals, param_var_mapping, allow_free_vars
        )
    else:
        # Parse an argument and check that it's valid for a `TypeParam`
        arg = arg_from_ast(node, globals, param_var_mapping, allow_free_vars)
        tyarg = _type_param.check_arg(arg, node)
        return tyarg.ty, InputFlags.NoFlags


def type_from_ast(
    node: AstNode,
    globals: Globals,
    param_var_mapping: dict[str, Parameter],
    allow_free_vars: bool = False,
) -> Type:
    """Turns an AST expression into a Guppy type."""
    ty, flags = type_with_flags_from_ast(
        node, globals, param_var_mapping, allow_free_vars
    )
    if flags != InputFlags.NoFlags:
        assert InputFlags.Inout not in flags  # Users shouldn't be able to set this
        raise GuppyError(FlagNotAllowedError(node))
    return ty


def type_row_from_ast(
    node: ast.expr, globals: "Globals", allow_free_vars: bool = False
) -> Sequence[Type]:
    """Turns an AST expression into a Guppy type row.

    This is needed to interpret the return type annotation of functions.
    """
    # The return type `-> None` is represented in the ast as `ast.Constant(value=None)`
    if isinstance(node, ast.Constant) and node.value is None:
        return []
    ty = type_from_ast(node, globals, {}, allow_free_vars)
    if isinstance(ty, TupleType):
        return ty.element_types
    else:
        return [ty]



from hugr import Wire
from hugr import tys as ht
from hugr.std.int import int_t

from guppylang.definition.custom import CustomInoutCallCompiler
from guppylang.definition.value import CallReturnWires
from guppylang.std._internal.compiler.arithmetic import inarrow_s, iwiden_s
from guppylang.std._internal.compiler.prelude import build_unwrap_right
from guppylang.std._internal.compiler.quantum import (
    QSYSTEM_RANDOM_EXTENSION,
    RNGCONTEXT_T,
)
from guppylang.std._internal.util import external_op


class RandomIntCompiler(CustomInoutCallCompiler):
    def compile_with_inouts(self, args: list[Wire]) -> CallReturnWires:
        [ctx] = args
        [rnd, ctx] = self.builder.add_op(
            external_op("RandomInt", [], ext=QSYSTEM_RANDOM_EXTENSION)(
                ht.FunctionType([RNGCONTEXT_T], [int_t(5), RNGCONTEXT_T]), []
            ),
            ctx,
        )
        [rnd] = self.builder.add_op(iwiden_s(5, 6), rnd)
        return CallReturnWires(regular_returns=[rnd], inout_returns=[ctx])


class RandomIntBoundedCompiler(CustomInoutCallCompiler):
    def compile_with_inouts(self, args: list[Wire]) -> CallReturnWires:
        [ctx, bound] = args
        bound_sum = self.builder.add_op(inarrow_s(6, 5), bound)
        bound = build_unwrap_right(
            self.builder, bound_sum, "bound must be a 32-bit integer"
        )
        [rnd, ctx] = self.builder.add_op(
            external_op("RandomIntBounded", [], ext=QSYSTEM_RANDOM_EXTENSION)(
                ht.FunctionType([RNGCONTEXT_T, int_t(5)], [int_t(5), RNGCONTEXT_T]), []
            ),
            ctx,
            bound,
        )
        [rnd] = self.builder.add_op(iwiden_s(5, 6), rnd)
        return CallReturnWires(regular_returns=[rnd], inout_returns=[ctx])



from guppylang.decorator import guppy
from guppylang.module import GuppyModule


def test_decorator():
    module = GuppyModule("test")

    @guppy
    def a() -> None:
        pass

    @guppy(module)
    def b() -> None:
        pass

    @guppy
    def c() -> None:
        pass

    default_module = guppy.get_module()

    assert not module.contains("a")
    assert module.contains("b")
    assert not module.contains("c")

    assert default_module.contains("a")
    assert not default_module.contains("b")
    assert default_module.contains("c")


def test_nested():
    def make_module() -> GuppyModule:
        @guppy
        def a() -> None:
            pass

        return guppy.get_module()

    module_a = make_module()
    module_b = make_module()

    assert module_a.contains("a")
    assert module_b.contains("a")



from guppylang.decorator import guppy
from guppylang.module import GuppyModule


module = GuppyModule("test")


def build():
    @guppy.struct(module)
    class Foo:
        x: "blah"


build()
module.compile()



import pathlib
import pytest

from tests.error.util import run_error_test

path = pathlib.Path(__file__).parent.resolve() / "linear_errors"
files = [
    x
    for x in path.iterdir()
    if x.is_file() and x.suffix == ".py" and x.name != "__init__.py"
]

# TODO: Skip functional tests for now
files = [f for f in files if "functional" not in f.name]

# Turn paths into strings, otherwise pytest doesn't display the names
files = [str(f) for f in files]


@pytest.mark.parametrize("file", files)
def test_linear_errors(file, capsys, snapshot):
    run_error_test(file, capsys, snapshot)




from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.builtins import array, owned
from guppylang.std.quantum import qubit

import guppylang.std.quantum_functional as quantum
from tests.util import compile_guppy


def test_basic_exec(validate, run_int_fn):
    module = GuppyModule("test")

    @guppy(module)
    def test() -> array[int, 10]:
        return array(i + 1 for i in range(10))

    @guppy(module)
    def main() -> int:
        s = 0
        for x in test():
            s += x
        return s

    package = module.compile()
    validate(package)
    run_int_fn(package, expected=sum(i + 1 for i in range(10)))


def test_basic_linear(validate):
    module = GuppyModule("test")
    module.load_all(quantum)
    module.load(qubit)

    @guppy(module)
    def test() -> array[qubit, 42]:
        return array(qubit() for _ in range(42))

    validate(module.compile())


def test_zero_length(validate, run_int_fn):
    module = GuppyModule("test")

    @guppy(module)
    def test() -> array[float, 0]:
        return array(i / 0 for i in range(0))

    @guppy(module)
    def main() -> int:
        test()
        return 0

    package = module.compile()
    validate(package)
    run_int_fn(package, expected=0)


def test_capture(validate, run_int_fn):
    module = GuppyModule("test")

    @guppy(module)
    def test(x: int) -> array[int, 42]:
        return array(i + x for i in range(42))

    @guppy(module)
    def main() -> int:
        s = 0
        for x in test(3):
            s += x
        return s

    package = module.compile()
    validate(package)
    run_int_fn(package, expected=sum(i + 3 for i in range(42)))


def test_capture_struct(validate):
    module = GuppyModule("test")

    @guppy.struct(module)
    class MyStruct:
        x: int
        y: float

    @guppy(module)
    def test(s: MyStruct) -> array[int, 42]:
        return array(i + s.x for i in range(42))

    validate(module.compile())


def test_scope(validate):
    @compile_guppy
    def test() -> float:
        x = 42.0
        array(x for x in range(10))
        return x

    validate(test)


def test_nested_left(validate, run_int_fn):
    module = GuppyModule("test")

    @guppy(module)
    def test() -> array[array[int, 10], 20]:
        return array(array(x + y for y in range(10)) for x in range(20))

    @guppy(module)
    def main() -> int:
        s = 0
        for xs in test():
            for x in xs:
                s += x
        return s

    package = module.compile()
    validate(package)
    run_int_fn(package, expected=sum(x + y for y in range(10) for x in range(20)))


def test_generic_size(validate):
    module = GuppyModule("test")
    n = guppy.nat_var("n", module)

    @guppy(module)
    def test(xs: array[int, n] @ owned) -> array[int, n]:
        return array(x + 1 for x in xs)

    validate(module.compile())


def test_generic_elem(validate):
    module = GuppyModule("test")
    T = guppy.type_var("T")

    @guppy(module)
    def foo(x: T) -> array[T, 10]:
        return array(x for _ in range(10))


def test_borrow(validate):
    module = GuppyModule("test")
    module.load_all(quantum)
    module.load(qubit)
    n = guppy.nat_var("n", module)

    @guppy.declare(module)
    def foo(q: qubit) -> int: ...

    @guppy(module)
    def test(q: qubit) -> array[int, n]:
        return array(foo(q) for _ in range(n))

    validate(module.compile())


def test_borrow_twice(validate):
    module = GuppyModule("test")
    module.load_all(quantum)
    module.load(qubit)
    n = guppy.nat_var("n", module)

    @guppy.declare(module)
    def foo(q: qubit) -> int: ...

    @guppy(module)
    def test(q: qubit) -> array[int, n]:
        return array(foo(q) + foo(q) for _ in range(n))

    validate(module.compile())


def test_borrow_struct(validate):
    module = GuppyModule("test")
    module.load_all(quantum)
    module.load(qubit)
    n = guppy.nat_var("n", module)

    @guppy.struct(module)
    class MyStruct:
        q1: qubit
        q2: qubit

    @guppy.declare(module)
    def foo(s: MyStruct) -> int: ...

    @guppy(module)
    def test(s: MyStruct) -> array[int, n]:
        return array(foo(s) for _ in range(n))

    validate(module.compile())


def test_borrow_and_consume(validate):
    module = GuppyModule("test")
    module.load_all(quantum)
    module.load(qubit)
    n = guppy.nat_var("n", module)

    @guppy.declare(module)
    def foo(q: qubit) -> int: ...

    @guppy.declare(module)
    def bar(q: qubit @ owned) -> int: ...

    @guppy(module)
    def test(qs: array[qubit, n] @ owned) -> array[int, n]:
        return array(foo(q) + bar(q) for q in qs)

    validate(module.compile())




from guppylang.decorator import guppy
from guppylang.module import GuppyModule


def test_docstring(validate):
    @guppy
    def f() -> None:
        "Example docstring"

    @guppy
    def g() -> None:
        """Multi
        line
        doc
        string.
        """
        f()

    @guppy
    def nested() -> None:
        """Test docstrings in nested functions"""

        def f_nested() -> None:
            "Example docstring"
            f_nested()
            g()

        def g_nested() -> None:
            """Multi
            line
            doc
            string.
            """

    default_module = guppy.get_module()
    validate(default_module.compile())


def test_docstring_module(validate):
    module = GuppyModule("")

    @guppy(module)
    def f() -> None:
        "Example docstring"
        f()

    @guppy(module)
    def g() -> None:
        """Multi
        line
        doc
        string.
        """

    @guppy(module)
    def nested() -> None:
        """Test docstrings in nested functions"""

        def f_nested() -> None:
            "Example docstring"
            f()
            g()

        def g_nested() -> None:
            """Multi
            line
            doc
            string.
            """
            f()
            f_nested()

    validate(module.compile())



import ast
import textwrap
from collections.abc import Callable, Mapping, Sequence
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Generic, Optional, TypeVar, cast

if TYPE_CHECKING:
    from guppylang.tys.ty import Type

AstNode = (
    ast.AST
    | ast.operator
    | ast.expr
    | ast.arg
    | ast.stmt
    | ast.Name
    | ast.keyword
    | ast.FunctionDef
)

T = TypeVar("T", covariant=True)


class AstVisitor(Generic[T]):
    """
    Note: This class is based on the implementation of `ast.NodeVisitor` but
    allows extra arguments to be passed to the `visit` functions.

    Original documentation:

    A node visitor base class that walks the abstract syntax tree and calls a
    visitor function for every node found.  This function may return a value
    which is forwarded by the `visit` method.

    This class is meant to be subclassed, with the subclass adding visitor
    methods.

    Per default the visitor functions for the nodes are ``'visit_'`` +
    class name of the node.  So a `TryFinally` node visit function would
    be `visit_TryFinally`.  This behavior can be changed by overriding
    the `visit` method.  If no visitor function exists for a node
    (return value `None`) the `generic_visit` visitor is used instead.

    Don't use the `NodeVisitor` if you want to apply changes to nodes during
    traversing.  For this a special visitor exists (`NodeTransformer`) that
    allows modifications.
    """

    def visit(self, node: Any, *args: Any, **kwargs: Any) -> T:
        """Visit a node."""
        method = "visit_" + node.__class__.__name__
        visitor = getattr(self, method, self.generic_visit)
        return visitor(node, *args, **kwargs)

    def generic_visit(self, node: Any, *args: Any, **kwargs: Any) -> T:
        """Called if no explicit visitor function exists for a node."""
        raise NotImplementedError(f"visit_{node.__class__.__name__} is not implemented")


class AstSearcher(ast.NodeVisitor):
    """Visitor that searches for occurrences of specific nodes in an AST."""

    matcher: Callable[[ast.AST], bool]
    dont_recurse_into: set[type[ast.AST]]
    found: list[ast.AST]
    is_first_node: bool

    def __init__(
        self,
        matcher: Callable[[ast.AST], bool],
        dont_recurse_into: set[type[ast.AST]] | None = None,
    ) -> None:
        self.matcher = matcher
        self.dont_recurse_into = dont_recurse_into or set()
        self.found = []
        self.is_first_node = True

    def generic_visit(self, node: ast.AST) -> None:
        if self.matcher(node):
            self.found.append(node)
        if self.is_first_node or type(node) not in self.dont_recurse_into:
            self.is_first_node = False
            super().generic_visit(node)


def find_nodes(
    matcher: Callable[[ast.AST], bool],
    node: ast.AST,
    dont_recurse_into: set[type[ast.AST]] | None = None,
) -> list[ast.AST]:
    """Returns all nodes in the AST that satisfy the matcher."""
    v = AstSearcher(matcher, dont_recurse_into)
    v.visit(node)
    return v.found


def name_nodes_in_ast(node: Any) -> list[ast.Name]:
    """Returns all `Name` nodes occurring in an AST."""
    found = find_nodes(lambda n: isinstance(n, ast.Name), node)
    return cast(list[ast.Name], found)


def return_nodes_in_ast(node: Any) -> list[ast.Return]:
    """Returns all `Return` nodes occurring in an AST."""
    found = find_nodes(lambda n: isinstance(n, ast.Return), node, {ast.FunctionDef})
    return cast(list[ast.Return], found)


def breaks_in_loop(node: Any) -> list[ast.Break]:
    """Returns all `Break` nodes occurring in a loop.

    Note that breaks in nested loops are excluded.
    """
    found = find_nodes(
        lambda n: isinstance(n, ast.Break), node, {ast.For, ast.While, ast.FunctionDef}
    )
    return cast(list[ast.Break], found)


class ContextAdjuster(ast.NodeTransformer):
    """Updates the `ast.Context` indicating if expressions occur on the LHS or RHS."""

    ctx: ast.expr_context

    def __init__(self, ctx: ast.expr_context) -> None:
        self.ctx = ctx

    def visit(self, node: ast.AST) -> ast.AST:
        return cast(ast.AST, super().visit(node))

    def visit_Name(self, node: ast.Name) -> ast.Name:
        return with_loc(node, ast.Name(id=node.id, ctx=self.ctx))

    def visit_Starred(self, node: ast.Starred) -> ast.Starred:
        return with_loc(node, ast.Starred(value=self.visit(node.value), ctx=self.ctx))

    def visit_Tuple(self, node: ast.Tuple) -> ast.Tuple:
        return with_loc(
            node, ast.Tuple(elts=[self.visit(elt) for elt in node.elts], ctx=self.ctx)
        )

    def visit_List(self, node: ast.List) -> ast.List:
        return with_loc(
            node, ast.List(elts=[self.visit(elt) for elt in node.elts], ctx=self.ctx)
        )

    def visit_Subscript(self, node: ast.Subscript) -> ast.Subscript:
        # Don't adjust the slice!
        return with_loc(
            node,
            ast.Subscript(value=self.visit(node.value), slice=node.slice, ctx=self.ctx),
        )

    def visit_Attribute(self, node: ast.Attribute) -> ast.Attribute:
        return with_loc(
            node,
            ast.Attribute(value=self.visit(node.value), attr=node.attr, ctx=self.ctx),
        )


@dataclass(frozen=True, eq=False)
class TemplateReplacer(ast.NodeTransformer):
    """Replaces nodes in a template."""

    replacements: Mapping[str, ast.AST | Sequence[ast.AST]]
    default_loc: ast.AST

    def _get_replacement(self, x: str) -> ast.AST | Sequence[ast.AST]:
        if x not in self.replacements:
            msg = f"No replacement for `{x}` is given"
            raise ValueError(msg)
        return self.replacements[x]

    def visit_Name(self, node: ast.Name) -> ast.AST:
        repl = self._get_replacement(node.id)
        if not isinstance(repl, ast.expr):
            msg = f"Replacement for `{node.id}` must be an expression"
            raise TypeError(msg)

        # Update the context
        adjuster = ContextAdjuster(node.ctx)
        return with_loc(repl, adjuster.visit(repl))

    def visit_Expr(self, node: ast.Expr) -> ast.AST | Sequence[ast.AST]:
        if isinstance(node.value, ast.Name):
            repl = self._get_replacement(node.value.id)
            repls = [repl] if not isinstance(repl, Sequence) else repl
            # Wrap expressions to turn them into statements
            return [
                with_loc(r, ast.Expr(value=r)) if isinstance(r, ast.expr) else r
                for r in repls
            ]
        return self.generic_visit(node)

    def generic_visit(self, node: ast.AST) -> ast.AST:
        # Insert the default location
        node = super().generic_visit(node)
        return with_loc(self.default_loc, node)


def template_replace(
    template: str, default_loc: ast.AST, **kwargs: ast.AST | Sequence[ast.AST]
) -> list[ast.stmt]:
    """Turns a template into a proper AST by substituting all placeholders."""
    nodes = ast.parse(textwrap.dedent(template)).body
    replacer = TemplateReplacer(kwargs, default_loc)
    new_nodes = []
    for n in nodes:
        new = replacer.visit(n)
        if isinstance(new, list):
            new_nodes.extend(new)
        else:
            new_nodes.append(new)
    return new_nodes


def line_col(node: ast.AST) -> tuple[int, int]:
    """Returns the line and column of an ast node."""
    return node.lineno, node.col_offset


def set_location_from(node: ast.AST, loc: ast.AST) -> None:
    """Copy source location from one AST node to the other."""
    node.lineno = loc.lineno
    node.col_offset = loc.col_offset
    node.end_lineno = loc.end_lineno
    node.end_col_offset = loc.end_col_offset

    source, file, line_offset = get_source(loc), get_file(loc), get_line_offset(loc)
    assert source is not None
    assert file is not None
    assert line_offset is not None
    annotate_location(node, source, file, line_offset)


def annotate_location(
    node: ast.AST, source: str, file: str, line_offset: int, recurse: bool = True
) -> None:
    node.line_offset = line_offset  # type: ignore[attr-defined]
    node.file = file  # type: ignore[attr-defined]
    node.source = source  # type: ignore[attr-defined]

    if recurse:
        for _field, value in ast.iter_fields(node):
            if isinstance(value, list):
                for item in value:
                    if isinstance(item, ast.AST):
                        annotate_location(item, source, file, line_offset, recurse)
            elif isinstance(value, ast.AST):
                annotate_location(value, source, file, line_offset, recurse)


def shift_loc(node: ast.AST, delta_lineno: int, delta_col_offset: int) -> None:
    """Shifts all line and column number in the AST node by the given amount."""
    if hasattr(node, "lineno"):
        node.lineno += delta_lineno
    if hasattr(node, "end_lineno") and node.end_lineno is not None:
        node.end_lineno += delta_lineno
    if hasattr(node, "col_offset"):
        node.col_offset += delta_col_offset
    if hasattr(node, "end_col_offset") and node.end_col_offset is not None:
        node.end_col_offset += delta_col_offset
    for _, value in ast.iter_fields(node):
        if isinstance(value, list):
            for item in value:
                if isinstance(item, ast.AST):
                    shift_loc(item, delta_lineno, delta_col_offset)
        elif isinstance(value, ast.AST):
            shift_loc(value, delta_lineno, delta_col_offset)


def get_file(node: AstNode) -> str | None:
    """Tries to retrieve a file annotation from an AST node."""
    try:
        file = node.file  # type: ignore[union-attr]
        return file if isinstance(file, str) else None
    except AttributeError:
        return None


def get_source(node: AstNode) -> str | None:
    """Tries to retrieve a source annotation from an AST node."""
    try:
        source = node.source  # type: ignore[union-attr]
        return source if isinstance(source, str) else None
    except AttributeError:
        return None


def get_line_offset(node: AstNode) -> int | None:
    """Tries to retrieve a line offset annotation from an AST node."""
    try:
        line_offset = node.line_offset  # type: ignore[union-attr]
        return line_offset if isinstance(line_offset, int) else None
    except AttributeError:
        return None


A = TypeVar("A", bound=ast.AST)


def with_loc(loc: ast.AST, node: A) -> A:
    """Copy source location from one AST node to the other."""
    set_location_from(node, loc)
    return node


def with_type(ty: "Type", node: A) -> A:
    """Annotates an AST node with a type."""
    node.type = ty  # type: ignore[attr-defined]
    return node


def get_type_opt(node: AstNode) -> Optional["Type"]:
    """Tries to retrieve a type annotation from an AST node."""
    from guppylang.tys.ty import Type, TypeBase

    try:
        ty = node.type  # type: ignore[union-attr]
        return cast(Type, ty) if isinstance(ty, TypeBase) else None
    except AttributeError:
        return None


def get_type(node: AstNode) -> "Type":
    """Retrieve a type annotation from an AST node.

    Fails if the node is not annotated.
    """
    ty = get_type_opt(node)
    assert ty is not None
    return ty


def has_empty_body(func_ast: ast.FunctionDef) -> bool:
    """Returns `True` if the body of a function definition is empty.

    This is the case if the body only contains a single `pass` statement or an ellipsis
    `...` expression.
    """
    if len(func_ast.body) == 0:
        return True
    if len(func_ast.body) > 1:
        return False
    [n] = func_ast.body
    return (
        isinstance(n, ast.Expr)
        and isinstance(n.value, ast.Constant)
        and n.value.value == Ellipsis
    )



import functools
from collections.abc import Sequence

from hugr import Wire, ops
from hugr import tys as ht
from hugr.build import cfg as hc
from hugr.build.dfg import DP, DfBase
from hugr.hugr.node_port import ToNode

from guppylang.checker.cfg_checker import CheckedBB, CheckedCFG, Row, Signature
from guppylang.checker.core import Place, Variable
from guppylang.compiler.core import (
    CompilerContext,
    DFContainer,
    is_return_var,
    return_var,
)
from guppylang.compiler.expr_compiler import ExprCompiler
from guppylang.compiler.stmt_compiler import StmtCompiler
from guppylang.tys.ty import SumType, row_to_type, type_to_row


def compile_cfg(
    cfg: CheckedCFG[Place],
    container: DfBase[DP],
    inputs: Sequence[Wire],
    ctx: CompilerContext,
) -> hc.Cfg:
    """Compiles a CFG to Hugr."""
    # Patch the CFG with dummy return variables
    # TODO: This mutates the CFG in-place which leads to problems when trying to lower
    #  the same function to Hugr twice. For now we just check that the return vars
    #  haven't already been inserted, but we should figure out a better way to handle
    #  this: https://github.com/CQCL/guppylang/issues/428
    if all(
        not is_return_var(v.name)
        for v in cfg.exit_bb.sig.input_row
        if isinstance(v, Variable)
    ):
        insert_return_vars(cfg)

    builder = container.add_cfg(*inputs)

    # Explicitly annotate the output types since Hugr can't infer them if the exit is
    # unreachable
    out_tys = [place.ty.to_hugr() for place in cfg.exit_bb.sig.input_row]
    # TODO: Use proper API for this once it's added in hugr-py:
    #  https://github.com/CQCL/hugr/issues/1816
    builder._exit_op._cfg_outputs = out_tys
    builder.parent_op._outputs = out_tys
    builder.parent_node = builder.hugr._update_node_outs(
        builder.parent_node, len(out_tys)
    )

    blocks: dict[CheckedBB[Place], ToNode] = {}
    for bb in cfg.bbs:
        blocks[bb] = compile_bb(bb, builder, bb == cfg.entry_bb, ctx)
    for bb in cfg.bbs:
        for i, succ in enumerate(bb.successors):
            builder.branch(blocks[bb][i], blocks[succ])

    return builder


def compile_bb(
    bb: CheckedBB[Place],
    builder: hc.Cfg,
    is_entry: bool,
    ctx: CompilerContext,
) -> ToNode:
    """Compiles a single basic block to Hugr.

    If the basic block is the output block, returns `None`.
    """
    # The exit BB is completely empty
    if bb.is_exit:
        assert len(bb.statements) == 0
        return builder.exit

    # Unreachable BBs (besides the exit) should have been removed by now
    assert bb.reachable

    # Otherwise, we use a regular `Block` node
    block: hc.Block
    inputs: Sequence[Place]
    if is_entry:
        inputs = bb.sig.input_row
        block = builder.add_entry()
    else:
        inputs = sort_vars(bb.sig.input_row)
        block = builder.add_block(*(v.ty.to_hugr() for v in inputs))

    # Add input node and compile the statements
    dfg = DFContainer(block)
    for v, wire in zip(inputs, block.input_node, strict=True):
        dfg[v] = wire
    dfg = StmtCompiler(ctx).compile_stmts(bb.statements, dfg)

    # If we branch, we also have to compile the branch predicate
    if len(bb.successors) > 1:
        assert bb.branch_pred is not None
        branch_port = ExprCompiler(ctx).compile(bb.branch_pred, dfg)
    else:
        # Even if we don't branch, we still have to add a `Sum(())` predicates
        branch_port = dfg.builder.add_op(ops.Tag(0, ht.UnitSum(1)))

    # Finally, we have to add the block output.
    outputs: Sequence[Place]
    if len(bb.successors) == 1:
        # The easy case is if we don't branch: We just output all variables that are
        # specified by the signature
        [outputs] = bb.sig.output_rows
    else:
        # CFG building ensures that branching BBs don't branch to the exit (exit jumps
        # must always be unconditional)
        assert not any(succ.is_exit for succ in bb.successors)

        # If we branch and the branches use the same places, then we can use a
        # regular output
        first, *rest = bb.sig.output_rows
        if all({p.id for p in first} == {p.id for p in r} for r in rest):
            outputs = first
        else:
            # Otherwise, we have to output a TupleSum: We put all non-linear variables
            # into the branch TupleSum and all linear variables in the normal output
            # (since they are shared between all successors). This is in line with the
            # ordering on variables which puts linear variables at the end.
            # We don't need to worry about the order of return vars since this isn't
            # a branch to an exit (see assert above).
            branch_port = choose_vars_for_tuple_sum(
                unit_sum=branch_port,
                output_vars=[
                    [v for v in sort_vars(row) if not v.ty.linear]
                    for row in bb.sig.output_rows
                ],
                dfg=dfg,
            )
            outputs = [v for v in first if v.ty.linear]

    # If this is *not* a jump to the exit BB, we need to sort the outputs to make the
    # signature consistent with what the next BB expects
    if not any(succ.is_exit for succ in bb.successors):
        outputs = sort_vars(outputs)
    else:
        # Exit variables are not allowed to be sorted since their order corresponds to
        # the function outputs
        assert len(bb.successors) == 1, "Exit jumps are always unconditional"

    block.set_block_outputs(branch_port, *(dfg[v] for v in outputs))
    return block


def insert_return_vars(cfg: CheckedCFG[Place]) -> None:
    """Patches a CFG by annotating dummy return variables in the BB signatures.

    The statement compiler turns `return` statements into assignments of dummy variables
    `%ret0`, `%ret1`, etc. We update the exit BB signature to make sure they are
    correctly outputted.
    """
    return_vars = [
        Variable(return_var(i), ty, None)
        for i, ty in enumerate(type_to_row(cfg.output_ty))
    ]
    # Prepend return variables to the exit signature
    cfg.exit_bb.sig = Signature(
        [*return_vars, *cfg.exit_bb.sig.input_row], cfg.exit_bb.sig.output_rows
    )
    # Also patch the predecessors
    for pred in cfg.exit_bb.predecessors:
        # The exit BB will be the only successor
        assert len(pred.sig.output_rows) == 1
        [out_row] = pred.sig.output_rows
        pred.sig = Signature(pred.sig.input_row, [[*return_vars, *out_row]])


def choose_vars_for_tuple_sum(
    unit_sum: Wire, output_vars: list[Row[Place]], dfg: DFContainer
) -> Wire:
    """Selects an output based on a TupleSum.

    Given `unit_sum: Sum(*(), *(), ...)` and output variable rows `#s1, #s2, ...`,
    constructs a TupleSum value of type `Sum(#s1, #s2, ...)`.
    """
    assert all(not v.ty.linear for var_row in output_vars for v in var_row)
    tys = [[v.ty for v in var_row] for var_row in output_vars]
    sum_type = SumType([row_to_type(row) for row in tys]).to_hugr()

    with dfg.builder.add_conditional(unit_sum) as conditional:
        for i, var_row in enumerate(output_vars):
            with conditional.add_case(i) as case:
                tag = case.add_op(ops.Tag(i, sum_type), *(dfg[v] for v in var_row))
                case.set_outputs(tag)
        return conditional


def compare_var(p1: Place, p2: Place) -> int:
    """Defines a `<` order on variables.

    We use this to determine in which order variables are outputted from basic blocks.
    We need to output linear variables at the end, so we do a lexicographic ordering of
    linearity and name.
    """
    return -1 if (p1.ty.linear, str(p1)) < (p2.ty.linear, str(p2)) else 1


def sort_vars(row: Row[Place]) -> list[Place]:
    """Sorts a row of variables.

    This determines the order in which they are outputted from a BB.
    """
    return sorted(row, key=functools.cmp_to_key(compare_var))



from tests.util import compile_guppy


@compile_guppy
def foo(x: int) -> int:
    y = x + 1

    def bar() -> None:

        def baz() -> None:
            y += 2

        baz()

    bar()
    return y



from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from tests.util import compile_guppy


def test_basic_type(validate):
    @compile_guppy
    def foo(x: str) -> str:
        return x

    validate(foo)


def test_basic_value(validate):
    @compile_guppy
    def foo() -> str:
        x = "Hello World"
        return x

    validate(foo)


def test_struct(validate):
    module = GuppyModule("module")

    @guppy.struct(module)
    class StringStruct:
        x: str

    @guppy(module)
    def main(s: StringStruct) -> None:
        StringStruct("Lorem Ipsum")

    validate(module.compile())



"""Collection of error messages emitted during linearity checking."""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, ClassVar

from guppylang.diagnostic import Error, Help, Note

if TYPE_CHECKING:
    from guppylang.checker.core import (
        Place,
        Variable,
    )
    from guppylang.checker.linearity_checker import UseKind
    from guppylang.definition.struct import StructField
    from guppylang.tys.ty import (
        StructType,
        Type,
    )


@dataclass(frozen=True)
class AlreadyUsedError(Error):
    title: ClassVar[str] = "Copy violation"
    span_label: ClassVar[str] = (
        "{place.describe} with non-copyable type `{place.ty}` cannot be "
        "{kind.subjunctive} ..."
    )
    place: Place
    kind: UseKind

    @dataclass(frozen=True)
    class PrevUse(Note):
        span_label: ClassVar[str] = "since it was already {prev_kind.subjunctive} here"
        prev_kind: UseKind


@dataclass(frozen=True)
class ComprAlreadyUsedError(Error):
    title: ClassVar[str] = "Copy violation"
    span_label: ClassVar[str] = (
        "{place.describe} with non-copyable type `{place.ty}` would be "
        "{kind.subjunctive} multiple times when evaluating this comprehension"
    )
    place: Place
    kind: UseKind

    @dataclass(frozen=True)
    class PrevUse(Note):
        span_label: ClassVar[str] = "since it was already {prev_kind.subjunctive} here"
        prev_kind: UseKind


@dataclass(frozen=True)
class PlaceNotUsedError(Error):
    title: ClassVar[str] = "Drop violation"
    place: Place

    @property
    def rendered_span_label(self) -> str:
        s = f"{self.place.describe} with non-droppable type `{self.place.ty}` "
        match self.children:
            case [PlaceNotUsedError.Branch(), *_]:
                return s + "may be leaked ..."
            case _:
                return s + "is leaked"

    @dataclass(frozen=True)
    class Branch(Note):
        span_label: ClassVar[str] = "if this expression is `{truth_value}`"
        truth_value: bool

    @dataclass(frozen=True)
    class Fix(Help):
        message: ClassVar[str] = (
            "Make sure that `{place}` is consumed or returned to avoid the leak"
        )


@dataclass(frozen=True)
class UnnamedExprNotUsedError(Error):
    title: ClassVar[str] = "Drop violation"
    span_label: ClassVar[str] = "Expression with non-droppable type `{ty}` is leaked"
    ty: Type

    @dataclass(frozen=True)
    class Fix(Help):
        message: ClassVar[str] = "Consider assigning this value to a local variable"


@dataclass(frozen=True)
class UnnamedFieldNotUsedError(Error):
    title: ClassVar[str] = "Drop violation"
    span_label: ClassVar[str] = (
        "Non-droppable field `{field.name}` of expression with type `{struct_ty}` is "
        "leaked"
    )
    field: StructField
    struct_ty: StructType

    @dataclass(frozen=True)
    class Fix(Help):
        message: ClassVar[str] = (
            "Consider assigning this value to a local variable before accessing the "
            "field `{used_field.name}`"
        )
        used_field: StructField


@dataclass(frozen=True)
class UnnamedSubscriptNotUsedError(Error):
    title: ClassVar[str] = "Drop violation"
    span_label: ClassVar[str] = (
        "Non-droppable items of expression with type `{container_ty}` are leaked ..."
    )
    container_ty: Type

    @dataclass(frozen=True)
    class SubscriptHint(Note):
        span_label: ClassVar[str] = "since only this subscript is used"

    @dataclass(frozen=True)
    class Fix(Help):
        message: ClassVar[str] = (
            "Consider assigning this value to a local variable before subscripting it"
        )


@dataclass(frozen=True)
class NotOwnedError(Error):
    title: ClassVar[str] = "Not owned"
    place: Place
    kind: UseKind
    is_call_arg: bool
    func_name: str | None
    calling_func_name: str

    @property
    def rendered_span_label(self) -> str:
        if self.is_call_arg:
            f = f"Function `{self.func_name}`" if self.func_name else "Function"
            return (
                f"{f} wants to take ownership of this argument, but "
                f"`{self.calling_func_name}` doesn't own `{self.place}`"
            )
        return (
            f"Cannot {self.kind.indicative} `{self.place}` since "
            f"`{self.calling_func_name}` doesn't own it"
        )

    @dataclass(frozen=True)
    class MakeOwned(Help):
        span_label: ClassVar[str] = (
            "Argument `{place.root.name}` is only borrowed. Consider taking ownership: "
            "`{place.root.name}: {place.root.ty} @owned`"
        )


@dataclass(frozen=True)
class MoveOutOfSubscriptError(Error):
    title: ClassVar[str] = "Subscript {kind.subjunctive}"
    span_label: ClassVar[str] = (
        "Cannot {kind.indicative} a subscript of `{parent}` with non-copyable type "
        "`{parent.ty}`"
    )
    kind: UseKind
    parent: Place

    @dataclass(frozen=True)
    class Explanation(Note):
        message: ClassVar[str] = (
            "Subscripts on non-copyable types are only allowed to be borrowed, not "
            "{kind.subjunctive}"
        )


@dataclass(frozen=True)
class BorrowShadowedError(Error):
    title: ClassVar[str] = "Borrow shadowed"
    span_label: ClassVar[str] = "Assignment shadows borrowed argument `{place}`"
    place: Place

    @dataclass(frozen=True)
    class Rename(Help):
        message: ClassVar[str] = "Consider assigning to a different name"


@dataclass(frozen=True)
class BorrowSubPlaceUsedError(Error):
    title: ClassVar[str] = "Copy violation"
    span_label: ClassVar[str] = (
        "Borrowed argument {borrowed_var} cannot be returned to the caller ..."
    )
    borrowed_var: Variable
    sub_place: Place

    @dataclass(frozen=True)
    class PrevUse(Note):
        span_label: ClassVar[str] = (
            "since `{sub_place}` with non-copyable type `{sub_place.ty}` was already "
            "{kind.subjunctive} here"
        )
        kind: UseKind

    @dataclass(frozen=True)
    class Fix(Help):
        message: ClassVar[str] = (
            "Consider writing a value back into `{sub_place}` before returning"
        )


@dataclass(frozen=True)
class DropAfterCallError(Error):
    title: ClassVar[str] = "Drop violation"
    span_label: ClassVar[str] = (
        "Value with non-droppable type `{ty}` would be leaked after {func} returns"
    )
    ty: Type
    func_name: str | None

    @property
    def func(self) -> str:
        return f"`{self.func_name}`" if self.func_name else "the function"

    @dataclass(frozen=True)
    class Assign(Help):
        message: ClassVar[str] = (
            "Consider assigning the value to a local variable before passing it to "
            "{func}"
        )


@dataclass(frozen=True)
class NonCopyableCaptureError(Error):
    title: ClassVar[str] = "Copy violation"
    span_label: ClassVar[str] = (
        "{var.describe} with non-copyable type {var.ty} cannot be used here since "
        "`{var}` is captured from an outer scope"
    )
    var: Variable

    @dataclass(frozen=True)
    class DefinedHere(Note):
        span_label: ClassVar[str] = "`{var}` defined here"


@dataclass(frozen=True)
class NonCopyablePartialApplyError(Error):
    title: ClassVar[str] = "Copy violation"
    span_label: ClassVar[str] = (
        "This expression implicitly constructs a closure that captures a non-copyable "
        "value"
    )

    @dataclass(frozen=True)
    class Captured(Note):
        span_label: ClassVar[str] = (
            "This expression with non-copyable type `{ty}` is implicitly captured"
        )
        ty: Type


@dataclass(frozen=True)
class NonDroppableForBreakError(Error):
    title: ClassVar[str] = "Break in non-droppable loop"
    span_label: ClassVar[str] = "Early exit in non-droppable loops is not allowed"

    @dataclass(frozen=True)
    class NonDroppableIteratorType(Note):
        span_label: ClassVar[str] = "Iterator has non-droppable type `{ty}`"
        ty: Type



import functools
from collections.abc import Sequence
from typing import Any

from guppylang.error import InternalGuppyError
from guppylang.tys.arg import Argument, ConstArg, TypeArg
from guppylang.tys.common import Transformer
from guppylang.tys.const import BoundConstVar, Const, ConstBase, ExistentialConstVar
from guppylang.tys.ty import (
    BoundTypeVar,
    ExistentialTypeVar,
    FunctionType,
    Type,
    TypeBase,
)
from guppylang.tys.var import ExistentialVar

Subst = dict[ExistentialVar, Type | Const]
Inst = Sequence[Argument]


class Substituter(Transformer):
    """Type transformer that applies a substitution of existential variables."""

    def __init__(self, subst: Subst) -> None:
        self.subst = subst

    @functools.singledispatchmethod
    def transform(self, ty: Any) -> Any | None:  # type: ignore[override]
        return None

    @transform.register
    def _transform_ExistentialTypeVar(self, ty: ExistentialTypeVar) -> Type | None:
        s = self.subst.get(ty, None)
        assert not isinstance(s, ConstBase)
        return s

    @transform.register
    def _transform_ExistentialConstVar(self, c: ExistentialConstVar) -> Const | None:
        s = self.subst.get(c, None)
        assert not isinstance(s, TypeBase)
        return s


class Instantiator(Transformer):
    """Type transformer that instantiates bound variables."""

    def __init__(self, inst: Inst) -> None:
        self.inst = inst

    @functools.singledispatchmethod
    def transform(self, ty: Any) -> Any | None:  # type: ignore[override]
        return None

    @transform.register
    def _transform_BoundTypeVar(self, ty: BoundTypeVar) -> Type | None:
        # Instantiate if type for the index is available
        if ty.idx < len(self.inst):
            arg = self.inst[ty.idx]
            assert isinstance(arg, TypeArg)
            return arg.ty

        # Otherwise, lower the de Bruijn index
        return BoundTypeVar(
            ty.display_name, ty.idx - len(self.inst), ty.copyable, ty.droppable
        )

    @transform.register
    def _transform_BoundConstVar(self, c: BoundConstVar) -> Const | None:
        # Instantiate if const value for the index is available
        if c.idx < len(self.inst):
            arg = self.inst[c.idx]
            assert isinstance(arg, ConstArg)
            return arg.const

        # Otherwise, lower the de Bruijn index
        return BoundConstVar(c.ty, c.display_name, c.idx - len(self.inst))

    @transform.register
    def _transform_FunctionType(self, ty: FunctionType) -> Type | None:
        if ty.parametrized:
            raise InternalGuppyError("Tried to instantiate under binder")
        return None



from guppylang.decorator import guppy
from guppylang.module import GuppyModule


module = GuppyModule("test")


@guppy.struct(module)
class StructA:
    x: "list[StructB]"


@guppy.struct(module)
class StructB:
    y: "StructA"


module.compile()



from guppylang import GuppyModule, guppy
from tests.integration.modules.mod_a import f, mod_a, MyType

mod_c = GuppyModule("mod_c")
mod_c.load_all(mod_a)


@guppy.declare(mod_c)
def g() -> MyType:
    ...


@guppy(mod_c)
def h(x: int) -> int:
    return f(x)


# Extend type defined in module A
@guppy.extend_type(MyType, module=mod_c)
class _:
    @guppy(mod_c)
    def __int__(self: "MyType") -> int:
        return 0



"""All sorts of weird stuff is allowed when variables are not used."""

from tests.util import compile_guppy


def test_not_always_defined1(validate):
    @compile_guppy
    def foo(x: bool) -> int:
        if x:
            z = 5
        return 4

    validate(foo)


def test_not_always_defined2(validate):
    @compile_guppy
    def foo(x: bool) -> int:
        if x:
            pass
        else:
            z = 5
        return 4

    validate(foo)


def test_not_always_defined3(validate):
    @compile_guppy
    def foo(x: bool) -> int:
        if x:
            y = True
        else:
            z = 5
        return 4

    validate(foo)


def test_different_types1(validate):
    @compile_guppy
    def foo(x: bool) -> int:
        if x:
            z = True
        else:
            z = 5
        return 4

    validate(foo)


def test_different_types2(validate):
    @compile_guppy
    def foo(x: bool) -> int:
        z = False
        if x:
            z = True
        else:
            z = 5
        return 4

    validate(foo)


def test_different_types3(validate):
    @compile_guppy
    def foo(x: bool) -> int:
        z = False
        if x:
            z = True
        else:
            z = 5
        return 4

    validate(foo)


def test_while_change_type(validate):
    @compile_guppy
    def foo() -> None:
        x = 42
        while True:
            x = True

    validate(foo)


def test_if_expr_different_types(validate):
    @compile_guppy
    def foo(x: bool) -> None:
        5 if x else False

    validate(foo)



from guppylang.decorator import guppy
from guppylang.module import GuppyModule

module = GuppyModule("test")

T = guppy.type_var("T", module=module)

@guppy.declare(module)
def foo(x: T) -> T:
    ...

@guppy(module)
def main() -> int:
    return (foo, foo)(42, 42)

module.compile()



from guppylang.decorator import guppy
from guppylang.module import GuppyModule


module = GuppyModule("lib")


@guppy.struct(module)
class MyType:
    x: int


@guppy(module)
def foo(x: int) -> int:
    return x



from dataclasses import dataclass
from typing import ClassVar

from guppylang.diagnostic import Error


@dataclass(frozen=True)
class UnsupportedError(Error):
    title: ClassVar[str] = "Unsupported"
    span_label: ClassVar[str] = "{things} {is_are} not supported{extra}"
    things: str
    singular: bool = False
    unsupported_in: str = ""

    @property
    def is_are(self) -> str:
        return "is" if self.singular else "are"

    @property
    def extra(self) -> str:
        return f" in {self.unsupported_in}" if self.unsupported_in else ""


@dataclass(frozen=True)
class UnexpectedError(Error):
    title: ClassVar[str] = "Unexpected {things}"
    span_label: ClassVar[str] = "Unexpected {things}{extra}"
    things: str
    unexpected_in: str = ""

    @property
    def extra(self) -> str:
        return f" in {self.unexpected_in}" if self.unexpected_in else ""


@dataclass(frozen=True)
class ExpectedError(Error):
    title: ClassVar[str] = "Expected {things}"
    span_label: ClassVar[str] = "Expected {things}{extra}"
    things: str
    got: str = ""

    @property
    def extra(self) -> str:
        return f", got {self.got}" if self.got else ""



from typing import Callable

from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.builtins import owned


module = GuppyModule("test")


@guppy.declare(module)
def foo(f: Callable[[int @owned], None]) -> None: ...


module.compile()



from guppylang.decorator import guppy
from guppylang.module import GuppyModule


module = GuppyModule("test")


def build():
    @guppy(module)
    def foo(x: float) -> int:
        return x


build()
module.compile()



from pytket import Circuit

from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.quantum import qubit

circ = Circuit(2)
circ.X(0)
circ.Y(1)

module = GuppyModule("test")
module.load(qubit)

guppy.load_pytket("guppy_circ", circ, module)

@guppy(module)
def foo(q: qubit) -> None:
    guppy_circ(q)


module.compile()


from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.builtins import owned
from guppylang.std.quantum import qubit
from guppylang.std.quantum_functional import h

module = GuppyModule("test")
module.load(qubit, h)


@guppy(module)
def foo(q: qubit @owned) -> None:
    h(q)


module.compile()



from tests.util import compile_guppy


def test_infinite_loop(validate):
    @compile_guppy
    def foo() -> int:
        while True:
            pass

    validate(foo)


def test_counting_loop(validate):
    @compile_guppy
    def foo(i: int) -> int:
        while i > 0:
            i -= 1
        return 0

    validate(foo)


def test_break(validate):
    @compile_guppy
    def foo(i: int) -> int:
        while True:
            if i == 0:
                break
            i -= 1
        return 0

    validate(foo)


def test_continue(validate):
    @compile_guppy
    def foo(i: int) -> int:
        x = 42
        while True:
            if i % 2 == 0:
                continue
            x = x + i

    validate(foo)


def test_return_in_loop(validate):
    @compile_guppy
    def foo(i: int) -> int:
        x = 42
        while i > 0:
            if x >= 1337:
                return x + i
            x = x + i
            i -= 1
        return x

    validate(foo)


def test_nested_loop(validate):
    @compile_guppy
    def foo(x: int, y: int) -> int:
        p = 0
        while x > 0:
            s = 0
            while y > 0:
                s += x
                y -= 1
            p += s
            x -= 1
        return p

    validate(foo)


def test_nested_loop_break_continue(validate):
    @compile_guppy
    def foo(x: int, y: int) -> int:
        p = 0
        while x > 0:
            s = 0
            while True:
                if x % 2 == 0:
                    continue
                s += x
                if s > y:
                    s = y
                else:
                    break
                y -= 1
            p += s
            x -= 1
        return p

    validate(foo)



from guppylang.decorator import guppy
from guppylang.module import GuppyModule


module = GuppyModule("test1")


@guppy(module)
def foo(x: float) -> int:
    return x


# Declaring an implicit module will load the explicit one above since it is in
# scope, so we will get the type error even though none of the modules were explicitly
# compiled

@guppy
def bar() -> None:
    pass




import guppylang.std.quantum as quantum
from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.quantum import qubit
from guppylang.std.builtins import owned

module = GuppyModule("test")
module.load_all(quantum)


@guppy.declare(module)
def bar(q: qubit @owned) -> int: ...


@guppy.declare(module)
def baz(q: qubit) -> int: ...


@guppy(module)
def foo(qs: list[qubit] @owned) -> list[int]:
    return [baz(q) for q in qs if bar(q)]


module.compile()



from guppylang.decorator import guppy
from guppylang.std.builtins import range, SizedIter, Range, py
from guppylang.module import GuppyModule

def test_range(validate, run_int_fn):
    module = GuppyModule("test_range")

    @guppy(module)
    def main() -> int:
        total = 0
        for x in range(5):
            total += x + 100 # Make the initial 0 obvious
        return total

    @guppy(module)
    def negative() -> int:
        total = 0
        for x in range(-3):
            total += 100 + x
        return total

    @guppy(module)
    def non_static() -> int:
        total = 0
        n = 4
        for x in range(n + 1):
            total += x + 100  # Make the initial 0 obvious
        return total

    compiled = module.compile()
    validate(compiled)
    run_int_fn(compiled, expected=510)
    run_int_fn(compiled, expected=0, fn_name="negative")
    run_int_fn(compiled, expected=510, fn_name="non_static")


def test_static_size(validate):
    module = GuppyModule("test")

    @guppy(module)
    def negative() -> SizedIter[Range, 10]:
        return range(10)

    validate(module.compile())


def test_py_size(validate):
    module = GuppyModule("test")
    n = 10

    @guppy(module)
    def negative() -> SizedIter[Range, 10]:
        return range(py(n))

    validate(module.compile())


def test_static_generic_size(validate):
    module = GuppyModule("test")
    n = guppy.nat_var("n", module=module)

    @guppy(module)
    def negative() -> SizedIter[Range, n]:
        return range(n)

    validate(module.compile())




import guppylang.std.quantum as quantum
from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.builtins import array
from guppylang.std.quantum import qubit


module = GuppyModule("test")
module.load_all(quantum)

@guppy(module)
def main() -> None:
    qs = array(qubit() for _ in range(10)) 
    qs[0] = qubit()  # Leaks the old qubit at index 0

module.compile()


import pytest

from hugr import ops, val

from guppylang.decorator import guppy
from guppylang.module import GuppyModule


def test_extern_float(validate):
    module = GuppyModule("module")

    guppy.extern("ext", ty="float", module=module)

    @guppy(module)
    def main() -> float:
        return ext + ext  # noqa: F821

    package = module.compile()
    validate(package)

    hg = package.module
    [c] = [data.op for n, data in hg.nodes() if isinstance(data.op, ops.Const)]
    assert isinstance(c.val, val.Extension)
    assert c.val.val["symbol"] == "ext"


def test_extern_alt_symbol(validate):
    module = GuppyModule("module")

    guppy.extern("ext", ty="int", symbol="foo", module=module)

    @guppy(module)
    def main() -> int:
        return ext  # noqa: F821

    package = module.compile()
    validate(package)

    hg = package.module
    [c] = [data.op for n, data in hg.nodes() if isinstance(data.op, ops.Const)]
    assert isinstance(c.val, val.Extension)
    assert c.val.val["symbol"] == "foo"


def test_extern_tuple(validate):
    module = GuppyModule("module")

    guppy.extern("ext", ty="tuple[int, float]", module=module)

    @guppy(module)
    def main() -> float:
        x, y = ext  # noqa: F821
        return x + y

    validate(module.compile())


@pytest.mark.skip("See https://github.com/CQCL/guppylang/issues/827")
def test_extern_conditional_assign(validate):
    module = GuppyModule("module")

    guppy.extern("x", ty="int", module=module)

    @guppy(module)
    def main(b: bool) -> int:
        if b:
            x = 4
        return x

    validate(module.compile())



import string
import textwrap
from collections.abc import Mapping, Sequence
from dataclasses import dataclass, field
from enum import Enum, auto
from typing import (
    Any,
    ClassVar,
    Final,
    Literal,
    Protocol,
    overload,
    runtime_checkable,
)

from typing_extensions import Self

from guppylang.error import InternalGuppyError
from guppylang.span import Loc, SourceMap, Span, ToSpan, to_span


class DiagnosticLevel(Enum):
    """Severity levels for compiler diagnostics."""

    #: An error that makes it impossible to proceed, causing an immediate abort.
    FATAL = auto()

    #: A regular error that is encountered during compilation. This is the most common
    #: diagnostic case.
    ERROR = auto()

    #: A warning about the code being compiled. Doesn't prevent compilation from
    #: finishing.
    WARNING = auto()

    #: A message giving some additional context. Usually used as a sub-diagnostic of
    #: errors.
    NOTE = auto()

    #: A message suggesting how to fix something. Usually used as a sub-diagnostic of
    #: errors.
    HELP = auto()


@runtime_checkable
@dataclass(frozen=True)
class SubDiagnostic(Protocol):
    """A sub-diagnostic attached to a parent diagnostic.

    Can be used to give some additional context, for example a note attached to an
    error.
    """

    #: Severity level of the sub-diagnostic.
    level: ClassVar[DiagnosticLevel]

    #: Optional span of the source location associated with this sub-diagnostic.
    span: ToSpan | None

    #: Label that is printed next to the span highlight. Can only be used if a span is
    #: provided.
    span_label: ClassVar[str | None] = None

    #: Message that is printed if no span is provided.
    message: ClassVar[str | None] = None

    #: The parent main diagnostic this sub-diagnostic is attached to.
    _parent: "Diagnostic | None" = field(default=None, init=False)

    def __post_init__(self) -> None:
        if self.span_label and self.span is None:
            raise InternalGuppyError("SubDiagnostic: Span label provided without span")

    @property
    def rendered_message(self) -> str | None:
        """The message of this diagnostic with formatted placeholders if provided."""
        return self._render(self.message)

    @property
    def rendered_span_label(self) -> str | None:
        """The span label of this diagnostic with formatted placeholders if provided."""
        return self._render(self.span_label)

    @overload
    def _render(self, s: str) -> str: ...

    @overload
    def _render(self, s: None) -> None: ...

    def _render(self, s: str | None) -> str | None:
        """Helper method to fill in placeholder values in strings with fields of this
        diagnostic.
        """

        class CustomFormatter(string.Formatter):
            def get_value(
                _self, key: int | str, args: Sequence[Any], kwargs: Mapping[str, Any]
            ) -> Any:
                assert isinstance(key, str)
                if hasattr(self, key):
                    return getattr(self, key)
                return getattr(self._parent, key)

        return CustomFormatter().format(s) if s is not None else None


@runtime_checkable
@dataclass(frozen=True)
class Diagnostic(SubDiagnostic, Protocol):
    """Abstract base class for compiler diagnostics that are reported to users.

    These could be fatal errors, regular errors, or warnings (see `DiagnosticLevel`).
    """

    #: Short title for the diagnostic that is displayed at the top.
    title: ClassVar[str]

    #: Optional sub-diagnostics giving some additional context.
    children: list["SubDiagnostic"] = field(default_factory=list, init=False)

    def __post_init__(self) -> None:
        super().__post_init__()
        if self.span is None and self.children:
            raise InternalGuppyError(
                "Diagnostic: Span-less diagnostics can't have children (FIXME)"
            )

    @property
    def rendered_title(self) -> str:
        """The title of this diagnostic with formatted placeholders."""
        return self._render(self.title)

    def add_sub_diagnostic(self, sub: "SubDiagnostic") -> Self:
        """Adds a new sub-diagnostic."""
        if self.span and sub.span and to_span(sub.span).file != to_span(self.span).file:
            raise InternalGuppyError(
                "Diagnostic: Cross-file sub-diagnostics are not supported"
            )
        object.__setattr__(sub, "_parent", self)
        self.children.append(sub)
        return self

    @property
    def rendered_message(self) -> str | None:
        """The message of this diagnostic with formatted placeholders if provided."""
        return self._render(self.message)

    @property
    def rendered_span_label(self) -> str | None:
        """The span label of this diagnostic with formatted placeholders if provided."""
        return self._render(self.span_label)


@runtime_checkable
@dataclass(frozen=True)
class Fatal(Diagnostic, Protocol):
    """Compiler diagnostic for errors that makes it impossible to proceed, causing an
    immediate abort."""

    level: ClassVar[Literal[DiagnosticLevel.FATAL]] = DiagnosticLevel.FATAL


@runtime_checkable
@dataclass(frozen=True)
class Error(Diagnostic, Protocol):
    """Compiler diagnostic for regular errors that are encountered during
    compilation."""

    level: ClassVar[Literal[DiagnosticLevel.ERROR]] = DiagnosticLevel.ERROR


@runtime_checkable
@dataclass(frozen=True)
class Note(SubDiagnostic, Protocol):
    """Compiler sub-diagnostic giving some additional context."""

    level: ClassVar[Literal[DiagnosticLevel.NOTE]] = DiagnosticLevel.NOTE


@runtime_checkable
@dataclass(frozen=True)
class Help(SubDiagnostic, Protocol):
    """Compiler sub-diagnostic suggesting how to fix something."""

    level: ClassVar[Literal[DiagnosticLevel.HELP]] = DiagnosticLevel.HELP


class DiagnosticsRenderer:
    """Standard renderer for compiler diagnostics in human-readable format."""

    source: SourceMap
    buffer: list[str]

    #: Maximum amount of leading whitespace until we start trimming it
    MAX_LEADING_WHITESPACE: Final[int] = 12

    #: Amount of leading whitespace left after trimming for padding
    OPTIMAL_LEADING_WHITESPACE: Final[int] = 4

    #: Maximum length of span labels after which we insert a newline
    MAX_LABEL_LINE_LEN: Final[int] = 60

    #: Maximum length of messages after which we insert a newline
    MAX_MESSAGE_LINE_LEN: Final[int] = 80

    #: Number of preceding source lines we show to give additional context
    PREFIX_CONTEXT_LINES: Final[int] = 2

    def __init__(self, source: SourceMap) -> None:
        self.buffer = []
        self.source = source

    def render_diagnostic(self, diag: Diagnostic) -> None:
        """Renders a single diagnostic together with its sub-diagnostics.

        Example:

        ```
        Error: Short title for the diagnostic (at path/to/file.py:line:column)
           |
        42 | def foo(x: blah) -> None:
           |            ^^^^ Span label
           |
        55 |     x = bar() + baz
           |         ----- Sub-diagnostic label

        Longer message describing the error.

        note: Sub-diagnostic message without span
        ```
        """
        if diag.span is None:
            # Omit the title if we don't have a span, but a long message. This case
            # should be fairly rare.
            msg = diag.rendered_message or diag.rendered_title
            self.buffer += wrap(
                f"{self.level_str(diag.level)}: {msg}", self.MAX_MESSAGE_LINE_LEN
            )
        else:
            span = to_span(diag.span)
            level = self.level_str(diag.level)
            all_spans = [span] + [
                to_span(child.span) for child in diag.children if child.span
            ]
            max_lineno = max(s.end.line for s in all_spans)
            self.buffer.append(f"{level}: {diag.rendered_title} (at {span.start})")
            self.render_snippet(
                span,
                diag.rendered_span_label,
                max_lineno,
                is_primary=True,
                prefix_lines=self.PREFIX_CONTEXT_LINES,
            )
            # First render all sub-diagnostics that come with a span
            for sub_diag in diag.children:
                if sub_diag.span:
                    self.render_snippet(
                        to_span(sub_diag.span),
                        sub_diag.rendered_span_label,
                        max_lineno,
                        is_primary=False,
                    )
            if diag.rendered_message:
                self.buffer.append("")
                self.buffer += wrap(diag.rendered_message, self.MAX_MESSAGE_LINE_LEN)
        # Finally, render all sub-diagnostics that have a non-span message
        for sub_diag in diag.children:
            if sub_diag.rendered_message:
                self.buffer.append("")
                self.buffer += wrap(
                    f"{self.level_str(sub_diag.level)}: {sub_diag.rendered_message}",
                    self.MAX_MESSAGE_LINE_LEN,
                )

    def render_snippet(
        self,
        span: Span,
        label: str | None,
        max_lineno: int,
        is_primary: bool,
        prefix_lines: int = 0,
    ) -> None:
        """Renders the source associated with a span together with an optional label.

        ```
           |
        42 | def foo(x: blah) -> None:
           |            ^^^^ Span label. This could cover
           |                 multiple lines!
        ```

        Also supports spans covering multiple lines:

        ```
           |
        42 | def foo(x: int) -> None:
           | ^^^^^^^^^^^^^^^^^^^^^^^^
           | ...
        48 |     return bar()
           | ^^^^^^^^^^^^^^^^ Label covering the entire definition of foo
        ```

        If `is_primary` is `False`, the span is highlighted using `-` instead of `^`:

        ```
           |
        42 | def foo(x: blah) -> None:
           |            ---- Non-primary span label
        ```

        Optionally includes up to `prefix_lines` preceding source lines to give
        additional context.
        """
        # Check how much space we need to reserve for the leading line numbers
        ll_length = len(str(max_lineno))
        highlight_char = "^" if is_primary else "-"

        def render_line(line: str, line_number: int | None = None) -> None:
            """Helper method to render a line with the line number bar on the left."""
            ll = "" if line_number is None else str(line_number)
            self.buffer.append(" " * (ll_length - len(ll)) + ll + " | " + line)

        # One line of padding
        render_line("")

        # Grab all lines we want to display and remove excessive leading whitespace
        prefix_lines = min(prefix_lines, span.start.line - 1)
        all_lines = self.source.span_lines(span, prefix_lines)
        leading_whitespace = min(len(line) - len(line.lstrip()) for line in all_lines)
        if leading_whitespace > self.MAX_LEADING_WHITESPACE:
            remove = leading_whitespace - self.OPTIMAL_LEADING_WHITESPACE
            all_lines = [line[remove:] for line in all_lines]
            span = span.shift_left(remove)

        # Render prefix lines
        for i, line in enumerate(all_lines[:prefix_lines]):
            render_line(line, span.start.line - prefix_lines + i)
        span_lines = all_lines[prefix_lines:]

        if span.is_multiline:
            [first, *middle, last] = span_lines
            render_line(first, span.start.line)
            # Compute the subspan that only covers the first line and render its
            # highlight banner
            first_span = Span(span.start, Loc(span.file, span.start.line, len(first)))
            first_highlight = " " * first_span.start.column + highlight_char * len(
                first_span
            )
            render_line(first_highlight)
            # Omit everything in the middle
            if middle:
                render_line("...")
            # The last line is handled uniformly with the single-line case below.
            # Therefore, create a subspan that only covers the last line.
            last_span = Span(Loc(span.file, span.end.line, 0), span.end)
        else:
            [last] = span_lines
            last_span = span

        # Render the last span line and add highlights
        render_line(last, span.end.line)
        last_highlight = " " * last_span.start.column + highlight_char * len(last_span)

        # Render the label next to the highlight
        if label:
            [label_first, *label_rest] = wrap(
                label,
                self.MAX_LABEL_LINE_LEN,
                # One space after the last `^`
                initial_indent=" ",
                # Indent all subsequent lines to be aligned
                subsequent_indent=" " * (len(last_highlight) + 1),
            )
            render_line(last_highlight + label_first)
            for lbl in label_rest:
                render_line(lbl)
        else:
            render_line(last_highlight)

    @staticmethod
    def level_str(level: DiagnosticLevel) -> str:
        """Returns the text used to identify the different kinds of diagnostics."""
        return level.name.lower().capitalize()


def wrap(text: str, width: int, **kwargs: Any) -> list[str]:
    """Custom version of `textwrap.wrap` that correctly handles text with line breaks.

    Even with `replace_whitespace=False`, the original version doesn't count line breaks
    as new paragraphs and instead would produce bad results like

    ```
    short paragraph of text

    long paragraph of text xxxxx                             # broken too early :(
    xxxx xxxxxx xxxx xxxxx xxxxxxxx xxxxxxxxxxxx xxxxx xx
    xxxx xxxxxxx xxxxx xx xxxx xxxxxxx xxxx
    ```
    """
    return [
        line
        for paragraph in text.splitlines()
        for line in (textwrap.wrap(paragraph, width, **kwargs) if paragraph else [""])
    ]



from dataclasses import dataclass
from typing import ClassVar

from guppylang.diagnostic import Error, Help, Note
from guppylang.tys.ty import FunctionType


@dataclass(frozen=True)
class IllegalPyExpressionError(Error):
    title: ClassVar[str] = "Unsupported Python expression"
    span_label: ClassVar[str] = "Expression of type `{python_ty}` is not supported"
    python_ty: type


@dataclass(frozen=True)
class PyExprNotCPythonError(Error):
    title: ClassVar[str] = "Not running CPython"
    span_label: ClassVar[str] = (
        "Compile-time `py(...)` expressions are only supported in CPython"
    )


@dataclass(frozen=True)
class PyExprNotStaticError(Error):
    title: ClassVar[str] = "Not compile-time evaluatable"
    span_label: ClassVar[str] = (
        "Guppy variable `{guppy_var}` cannot be accessed in a compile-time `py(...)` "
        "expression"
    )
    guppy_var: str


@dataclass(frozen=True)
class PyExprEvalError(Error):
    title: ClassVar[str] = "Python error"
    span_label: ClassVar[str] = "Error occurred while evaluating this expression"
    message: ClassVar[str] = "Traceback printed below:\n\n{err}"
    err: str


@dataclass(frozen=True)
class PyExprIncoherentListError(Error):
    title: ClassVar[str] = "Unsupported list"
    span_label: ClassVar[str] = "List contains elements with different types"


@dataclass(frozen=True)
class Tket2NotInstalled(Error):
    title: ClassVar[str] = "Tket2 not installed"
    span_label: ClassVar[str] = (
        "Experimental pytket compatibility requires `tket2` to be installed"
    )

    @dataclass(frozen=True)
    class InstallInstruction(Help):
        message: ClassVar[str] = "Install tket2: `pip install tket2`"


@dataclass(frozen=True)
class PytketSignatureMismatch(Error):
    title: ClassVar[str] = "Signature mismatch"
    span_label: ClassVar[str] = "Signature `{name}` doesn't match provided circuit"
    name: str

    @dataclass(frozen=True)
    class TypeHint(Note):
        message: ClassVar[str] = "Expected circuit signature is `{circ_sig}`"
        circ_sig: FunctionType



from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.builtins import owned
from guppylang.std.quantum import qubit


module = GuppyModule("test")
module.load(qubit)


@guppy.declare(module)
def foo(x: list[qubit @owned]) -> qubit: ...


module.compile()



from guppylang.decorator import guppy
from guppylang.module import GuppyModule


module = GuppyModule("test")


@guppy.struct(module)
class MyStruct:
    x: int


@guppy(module)
def main() -> None:
    MyStruct(1, 2, 3)


module.compile()



"""Tests that the compiler correctly inserts order edges between operations that have
side-effects.
"""

from hugr import ops, Hugr, Node
from hugr.std import PRELUDE

from guppylang import guppy
from guppylang.module import GuppyModule
from guppylang.std._internal.compiler.quantum import RESULT_EXTENSION, QUANTUM_EXTENSION
from guppylang.std.builtins import result, array, owned, panic
from guppylang.std.quantum import qubit, discard, measure


def find_ext_nodes(hugr: Hugr, qualified_name: str) -> list[Node]:
    """Returns all extension nodes in a Hugr that match the given qualified name."""
    return [
        node for node, data in hugr.nodes() if name_matches(data.op, qualified_name)
    ]


def name_matches(op: ops.Op, qualified_name: str) -> bool:
    """Checks if an op is an extension op that matches the given qualified name."""
    match op:
        case ops.ExtOp() as ext_op:
            return ext_op.op_def().qualified_name() == qualified_name
        case ops.Custom(op_name=op_name, extension=extension):
            return qualified_name == (
                f"{extension}.{op_name}" if extension else op_name
            )
        case _:
            return False


def check_order(hugr: Hugr, nodes: list[Node]) -> None:
    """Checks that the provided nodes appear in the specified order in the order-edge
    graph."""
    # Do a DFS traversal of the order edge graph starting at the first node
    stack = [nodes[0]]
    visited = set()
    while stack:
        curr = stack.pop()
        visited.add(curr)
        if curr in nodes:
            # Check this node is the next one in the sequence
            assert curr == nodes.pop(0)
        for n in hugr.outgoing_order_links(curr):
            assert n not in visited, "Order edge graph must be acyclic"
            stack.append(n)
    # Check that all specified nodes occurred in the graph
    assert len(nodes) == 0


def test_result_panic(validate):
    module = GuppyModule("test")

    @guppy(module)
    def test() -> None:
        result("a", True)
        result("b", 10)
        panic("Boo!")
        result("c", 10.5)

    compiled = module.compile()
    validate(compiled)

    # Check that we have the expected order edges between the results
    hugr = compiled.module
    [a] = find_ext_nodes(hugr, RESULT_EXTENSION.get_op("result_bool").qualified_name())
    [b] = find_ext_nodes(hugr, RESULT_EXTENSION.get_op("result_int").qualified_name())
    [p] = find_ext_nodes(hugr, PRELUDE.get_op("panic").qualified_name())
    [c] = find_ext_nodes(hugr, RESULT_EXTENSION.get_op("result_f64").qualified_name())
    check_order(hugr, [a, b, p, c])


def test_qalloc_qfree(validate):
    module = GuppyModule("test")
    module.load(qubit, discard, measure)

    @guppy(module)
    def test() -> None:
        q1 = qubit()
        discard(q1)
        q2 = qubit()
        measure(q2)

    compiled = module.compile()
    validate(compiled)

    # Check that we have the expected order edges between the allocations and frees
    hugr = compiled.module
    [a1, a2] = find_ext_nodes(hugr, QUANTUM_EXTENSION.get_op("QAlloc").qualified_name())
    [d1] = find_ext_nodes(hugr, QUANTUM_EXTENSION.get_op("QFree").qualified_name())
    [d2] = find_ext_nodes(
        hugr, QUANTUM_EXTENSION.get_op("MeasureFree").qualified_name()
    )
    check_order(hugr, [a1, d1, a2, d2])


def test_call(validate):
    module = GuppyModule("test")
    module.load(qubit, discard)

    @guppy(module)
    def my_discard(q: qubit @ owned) -> None:
        discard(q)

    @guppy(module)
    def test() -> None:
        q1 = qubit()
        my_discard(q1)
        q2 = qubit()
        f = my_discard
        f(q2)

    compiled = module.compile()
    validate(compiled)

    # Check that we have the expected order edges between the allocations and calls
    hugr = compiled.module
    [a1, a2] = find_ext_nodes(hugr, QUANTUM_EXTENSION.get_op("QAlloc").qualified_name())
    [d1] = [node for node, data in hugr.nodes() if isinstance(data.op, ops.Call)]
    [d2] = [
        node for node, data in hugr.nodes() if isinstance(data.op, ops.CallIndirect)
    ]
    check_order(hugr, [a1, d1, a2, d2])


def test_nested(validate):
    module = GuppyModule("test")
    module.load(qubit, discard, measure)

    @guppy(module)
    def test() -> None:
        qs1 = array(qubit() for _ in range(10))
        array(measure(q) for q in qs1)
        qs2 = array(qubit() for _ in range(10))
        array(discard(q) for q in qs2)

    compiled = module.compile()
    validate(compiled)

    # Check that we have the expected order edges between the comprehension tail loops
    hugr = compiled.module
    [l1, l2, l3, l4] = [
        node for node, data in hugr.nodes() if isinstance(data.op, ops.TailLoop)
    ]
    check_order(hugr, [l1, l2, l3, l4])



from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.builtins import array, owned

from guppylang.std.quantum import qubit


def test_unpack_array(validate):
    module = GuppyModule("test")
    module.load(qubit)

    @guppy(module)
    def main(qs: array[qubit, 3] @ owned) -> tuple[qubit, qubit, qubit]:
        q1, q2, q3 = qs
        return q1, q2, q3

    validate(module.compile())


def test_unpack_starred(validate):
    module = GuppyModule("test")
    module.load(qubit)

    @guppy(module)
    def main(
        qs: array[qubit, 10] @ owned,
    ) -> tuple[qubit, qubit, qubit, qubit, qubit, qubit, array[qubit, 4]]:
        q1, q2, *qs, q3 = qs
        [q4, *qs] = qs
        *qs, q5, q6 = qs
        [*qs] = qs
        return q1, q2, q3, q4, q5, q6, qs

    validate(module.compile())


def test_unpack_starred_empty(validate):
    module = GuppyModule("test")
    module.load(qubit)

    @guppy(module)
    def main(qs: array[qubit, 2] @ owned) -> tuple[qubit, array[qubit, 0], qubit]:
        q1, *empty, q2 = qs
        return q1, empty, q2

    validate(module.compile())


def test_unpack_big_iterable(validate):
    # Test that the compile-time doesn't scale with the size of the unpacked iterable
    module = GuppyModule("test")
    module.load(qubit)

    @guppy(module)
    def main(qs: array[qubit, 1000] @ owned) -> tuple[qubit, array[qubit, 998], qubit]:
        q1, *qs, q2 = qs
        return q1, qs, q2

    validate(module.compile())


def test_unpack_range(validate, run_int_fn):
    module = GuppyModule("test")

    @guppy(module)
    def main() -> int:
        [_, x, *_, y, _] = range(10)
        return x + y

    compiled = module.compile()
    validate(compiled)
    run_int_fn(compiled, expected=9)


def test_unpack_tuple_starred(validate, run_int_fn):
    module = GuppyModule("test")

    @guppy(module)
    def main() -> array[int, 2]:
        x, *ys, z = 1, 2, 3, 4
        return ys

    validate(module.compile())


def test_unpack_nested(validate, run_int_fn):
    module = GuppyModule("test")

    @guppy(module)
    def main(
        xs: array[array[array[int, 5], 10], 20] @ owned,
    ) -> tuple[
        array[int, 5],  # x
        int,  # y
        array[int, 3],  # z
        array[array[int, 5], 8],  # a
        array[array[array[int, 5], 10], 18],  # b
        array[array[int, 5], 10],  # c
    ]:
        (x, [y, *z, _], *a), *b, c = xs
        return x, y, z, a, b, c

    validate(module.compile())


def test_left_to_right(validate, run_int_fn):
    module = GuppyModule("test")

    @guppy(module)
    def left() -> int:
        [x, x, *_] = range(10)
        return x

    @guppy(module)
    def right() -> int:
        [*_, x, x] = range(10)
        return x

    compiled = module.compile()
    validate(compiled)
    run_int_fn(compiled, fn_name="left", expected=1)
    run_int_fn(compiled, fn_name="right", expected=9)



from guppylang.decorator import guppy
from guppylang.module import GuppyModule


module = GuppyModule("test")

T = guppy.type_var("T", module=module)


@guppy.declare(module)
def foo(x: T) -> None:
    ...


@guppy(module)
def main(x: float) -> None:
    f = foo[int]
    f(x)


module.compile()



import pytest
from guppylang import qubit, guppy, GuppyModule
from guppylang.std import quantum
from guppylang.std.angles import angle
from guppylang.std.builtins import owned
from guppylang.std.quantum import cx, rz
from guppylang.std.quantum_functional import quantum_functional, h

from tests.util import compile_guppy


def test_types(validate):
    @compile_guppy
    def test(
        xs: list[int], ys: list[tuple[int, float]]
    ) -> tuple[list[int], list[tuple[int, float]]]:
        return xs, ys

    validate(test)


def test_len(validate):
    @compile_guppy
    def test(xs: list[int]) -> int:
        return len(xs)

    validate(test)


def test_literal(validate):
    @compile_guppy
    def test(x: float) -> list[float]:
        return [1.0, 2.0, 3.0, 4.0 + x]

    validate(test)


def test_literal_linear(validate):
    module = GuppyModule("test")
    module.load_all(quantum_functional)
    module.load(qubit)

    @guppy(module)
    def test(q1: qubit @owned, q2: qubit @owned) -> list[qubit]:
        return [q1, h(q2)]

    validate(module.compile())


def test_push_pop(validate):
    @compile_guppy
    def test(xs: list[int]) -> bool:
        xs.append(3)
        x = xs.pop()
        return x == 3

    validate(test)


@pytest.mark.skip("See https://github.com/CQCL/guppylang/issues/528")
def test_arith(validate):
    @compile_guppy
    def test(xs: list[int]) -> list[int]:
        xs += xs + [42]
        xs = 3 * xs
        return xs * 4

    validate(test)


@pytest.mark.skip("See https://github.com/CQCL/guppylang/issues/528")
def test_arith_linear(validate):
    module = GuppyModule("test")
    module.load_all(quantum_functional)
    module.load(qubit)

    @guppy(module)
    def test(xs: list[qubit] @owned, ys: list[qubit] @owned, q: qubit @owned) -> list[qubit]:
        xs += [q]
        return xs + ys

    validate(module.compile())


def test_subscript(validate):
    @compile_guppy
    def test(xs: list[float], i: int) -> float:
        return xs[2 * i]

    validate(test)


def test_linear(validate):
    module = GuppyModule("test")
    module.load(qubit)

    @guppy(module)
    def test(xs: list[qubit], q: qubit @owned) -> int:
        xs.append(q)
        return len(xs)

    validate(module.compile())


def test_subscript_drop_rest(validate):
    module = GuppyModule("test")
    module.load_all(quantum)

    @guppy.declare(module)
    def foo() -> list[int]: ...

    @guppy(module)
    def main() -> int:
        return foo()[0]

    validate(module.compile())


def test_linear_subscript(validate):
    module = GuppyModule("test")
    module.load_all(quantum)

    @guppy.declare(module)
    def foo(q: qubit) -> None: ...

    @guppy(module)
    def main(qs: list[qubit] @owned, i: int) -> list[qubit]:
        foo(qs[i])
        return qs

    validate(module.compile())


def test_inout_subscript(validate):
    module = GuppyModule("test")
    module.load_all(quantum)

    @guppy.declare(module)
    def foo(q: qubit) -> None: ...

    @guppy(module)
    def main(qs: list[qubit], i: int) -> None:
        foo(qs[i])

    validate(module.compile())


def test_multi_subscripts(validate):
    module = GuppyModule("test")
    module.load_all(quantum)

    @guppy.declare(module)
    def foo(q1: qubit, q2: qubit) -> None: ...

    @guppy(module)
    def main(qs: list[qubit] @owned) -> list[qubit]:
        foo(qs[0], qs[1])
        foo(qs[0], qs[0])  # Will panic at runtime
        return qs

    validate(module.compile())


def test_struct_list(validate):
    module = GuppyModule("test")
    module.load_all(quantum)

    @guppy.struct(module)
    class S:
        q1: qubit
        q2: qubit

    @guppy.declare(module)
    def foo(q1: qubit, q2: qubit) -> None: ...

    @guppy(module)
    def main(ss: list[S] @owned) -> list[S]:
        # This will panic at runtime :(
        # To make this work, we would need to replace the qubits in the struct
        # with `qubit | None` and write back `None` after `q1` has been extracted...
        foo(ss[0].q1, ss[0].q2)
        return ss

    validate(module.compile())


def test_nested_subscripts(validate):
    module = GuppyModule("test")
    module.load_all(quantum)

    @guppy.declare(module)
    def foo(q: qubit) -> None: ...

    @guppy.declare(module)
    def bar(
        q1: qubit, q2: qubit, q3: qubit, q4: qubit
    ) -> None: ...

    @guppy(module)
    def main(qs: list[list[qubit]] @owned) -> list[list[qubit]]:
        foo(qs[0][0])
        # The following should work *without* panicking at runtime! Accessing `qs[0][0]`
        # replaces one qubit with `None` but puts everything back into `qs` before
        # going to the next argument.
        bar(qs[0][0], qs[0][1], qs[1][0], qs[1][1])
        return qs

    validate(module.compile())


def test_struct_nested_subscript(validate):
    module = GuppyModule("test")
    module.load_all(quantum)

    @guppy.struct(module)
    class C:
        c: qubit
        blah: int

    @guppy.struct(module)
    class B:
        ys: list[list[C]]
        foo: C

    @guppy.struct(module)
    class A:
        xs: list[B]
        bar: qubit
        baz: tuple[B, B]

    @guppy.declare(module)
    def foo(q1: qubit) -> None: ...

    @guppy(module)
    def main(a: A @owned, i: int, j: int, k: int) -> A:
        foo(a.xs[i].ys[j][k].c)
        return a

    validate(module.compile())


def test_phase_gadget(validate):
    module = GuppyModule("test")
    module.load_all(quantum)
    module.load(angle)

    @guppy(module)
    def paulig(qs: list[qubit], alpha: angle) -> None:
        n = len(qs)
        for i in range(n - 1):
            cx(qs[i], qs[i + 1])
        rz(qs[n - 1], alpha)
        for i in range(n - 1):
            cx(qs[n - i - 1], qs[n - i - 2])

    validate(module.compile())



import numpy as np

from guppylang.decorator import guppy
from guppylang.std.angles import angle, pi
from guppylang.std.builtins import array, owned, py
from guppylang.std.option import Option, nothing, some
from guppylang.std.quantum import (
    cz,
    discard,
    h,
    measure,
    qubit,
    ry,
    rz,
)

phi = np.arccos(1 / 3)


# Preparation of approximate T state, from https://arxiv.org/abs/2310.12106
@guppy
def prepare_approx() -> qubit:
    q = qubit()
    ry(q, angle(py(phi)))
    rz(q, pi / 4)
    return q


# The inverse of the [[5,3,1]] encoder in figure 3 of https://arxiv.org/abs/2208.01863
@guppy
def distill(
    target: qubit @ owned,
    qs: array[qubit, 4] @ owned,
) -> tuple[qubit, bool]:
    """First argument is the target qubit which will be returned from the circuit.
    Other arguments are ancillae, which should also be in an approximate T state.
    Returns target qubit and a bool, which is true if the distillation succeeded.
    """
    cz(qs[0], qs[1])
    cz(qs[2], qs[3])
    cz(target, qs[0])
    cz(qs[1], qs[2])
    cz(target, qs[3])
    # Measuring gives false for success, true for failure.
    # We check for all falses to say whether distillation succeeded.
    for i in range(4):
        h(qs[i])
    bits = array(not measure(q) for q in qs)
    # guppy doesn't yet support the `any` or `all` operators...
    success = True
    for b in bits:
        success &= b
    return target, success


@guppy
def t_state(timeout: int) -> Option[qubit]:
    """Create a T state using magic state distillation with `timeout` attempts.

    On success returns a qubit in a magic T state.

    On failure (i.e. number of attempts are exceeded) returns nothing.
    """
    if timeout > 0:
        tgt = prepare_approx()
        qs = array(prepare_approx() for _ in range(4))

        q, success = distill(tgt, qs)
        if success:
            return some(q)
        else:
            # Discard the qubit and start over
            # Note, this could just as easily be a while loop!
            discard(q)
            return t_state(timeout - 1)

    # We ran out of attempts
    return nothing()


hugr = guppy.compile_module()



from collections import deque
from collections.abc import Iterator
from typing import Generic, TypeVar

from guppylang.cfg.analysis import (
    AssignmentAnalysis,
    DefAssignmentDomain,
    LivenessAnalysis,
    LivenessDomain,
    MaybeAssignmentDomain,
    Result,
)
from guppylang.cfg.bb import BB, BBStatement, VariableStats
from guppylang.nodes import InoutReturnSentinel

T = TypeVar("T", bound=BB)


class BaseCFG(Generic[T]):
    """Abstract base class for control-flow graphs."""

    bbs: list[T]
    entry_bb: T
    exit_bb: T

    live_before: Result[LivenessDomain[str]]
    ass_before: Result[DefAssignmentDomain[str]]
    maybe_ass_before: Result[MaybeAssignmentDomain[str]]

    def __init__(
        self, bbs: list[T], entry_bb: T | None = None, exit_bb: T | None = None
    ):
        self.bbs = bbs
        if entry_bb:
            self.entry_bb = entry_bb
        if exit_bb:
            self.exit_bb = exit_bb
        self.live_before = {}
        self.ass_before = {}
        self.maybe_ass_before = {}

    def ancestors(self, *bbs: T) -> Iterator[T]:
        """Returns an iterator over all ancestors of the given BBs in BFS order."""
        queue = deque(bbs)
        visited = set()
        while queue:
            bb = queue.popleft()
            if bb in visited:
                continue
            visited.add(bb)
            yield bb
            queue += bb.predecessors

    def update_reachable(self) -> None:
        """Sets the reachability flags on the BBs in this CFG."""
        queue = {self.entry_bb}
        while queue:
            bb = queue.pop()
            if not bb.reachable:
                bb.reachable = True
                for succ in bb.successors:
                    queue.add(succ)


class CFG(BaseCFG[BB]):
    """A control-flow graph of unchecked basic blocks."""

    def __init__(self) -> None:
        super().__init__([])
        self.entry_bb = self.new_bb()
        self.exit_bb = self.new_bb()

    def new_bb(self, *preds: BB, statements: list[BBStatement] | None = None) -> BB:
        """Adds a new basic block to the CFG."""
        bb = BB(
            len(self.bbs), self, predecessors=list(preds), statements=statements or []
        )
        self.bbs.append(bb)
        for p in preds:
            p.successors.append(bb)
        return bb

    def link(self, src_bb: BB, tgt_bb: BB) -> None:
        """Adds a control-flow edge between two basic blocks."""
        src_bb.successors.append(tgt_bb)
        tgt_bb.predecessors.append(src_bb)

    def dummy_link(self, src_bb: BB, tgt_bb: BB) -> None:
        """Adds a dummy control-flow edge between two basic blocks that is provably
        never taken.

        For example, a `if False: ...` statement emits such a dummy link.
        """
        src_bb.dummy_successors.append(tgt_bb)
        tgt_bb.dummy_predecessors.append(src_bb)

    def analyze(
        self,
        def_ass_before: set[str],
        maybe_ass_before: set[str],
        inout_vars: list[str],
    ) -> dict[BB, VariableStats[str]]:
        stats = {bb: bb.compute_variable_stats() for bb in self.bbs}
        # Mark all borrowed variables as implicitly used in the exit BB
        stats[self.exit_bb].used |= {x: InoutReturnSentinel(var=x) for x in inout_vars}
        # This also means borrowed variables are always live, so we can use them as the
        # initial value in the liveness analysis. This solves the edge case that
        # borrowed variables should be considered live, even if the exit is actually
        # unreachable (to avoid linearity violations later).
        inout_live = {x: self.exit_bb for x in inout_vars}
        self.live_before = LivenessAnalysis(
            stats, initial=inout_live, include_unreachable=True
        ).run(self.bbs)
        self.ass_before, self.maybe_ass_before = AssignmentAnalysis(
            stats, def_ass_before, maybe_ass_before, include_unreachable=True
        ).run_unpacked(self.bbs)
        return stats



from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.quantum import qubit

import guppylang.std.quantum as quantum

module = GuppyModule("test")
module.load_all(quantum)


T = guppy.type_var("T", module=module)


@guppy.declare(module)
def foo(x: T) -> None:
    ...


@guppy(module)
def main(q: qubit) -> None:
    foo(q)


module.compile()
